import pandas as pd
from typing import Iterable, Dict
import requests
import pyterrier as pt


TERRIER_DATA_BASE="http://data.terrier.org/indices/"
STANDARD_TERRIER_INDEX_FILES = [
    "data.direct.bf",
    "data.document.fsarrayfile",
    "data.inverted.bf",
    "data.lexicon.fsomapfile",
    "data.lexicon.fsomaphash",
    "data.lexicon.fsomapid",
    "data.meta.idx",
    "data.meta.zdata",
    "data.properties"
]


def passage_generate(dataset):
    for filename in dataset.get_corpus():
        with pt.io.autoopen(filename, 'rt') as corpusfile:
            for line in corpusfile: #for each line
                docno, passage = line.split("\t")
                yield {'docno' : docno, 'text' : passage}

def _datarepo_index(self, component, variant=None, version='latest', **kwargs):
    if variant is None:
        raise ValueError(f"Must specify index variant for {self.name}. See http://data.terrier.org/{self.name}.dataset.html")
    urlprefix= f"http://data.terrier.org/indices/{self.name}/{variant}/{version}/"
    url = urlprefix + "files"
    try:
        r = requests.get(url, **kwargs)
        r.raise_for_status()
        file = r.text.splitlines()
    except Exception as e:
        raise ValueError(f"Could not find index variant {variant} for dataset {self.name} at {url}. See available variants at http://data.terrier.org/{self.name}.dataset.html") from e
    rtr = []
    import re
    for linenum, line in enumerate(file):
        # skip comments
        if line.startswith("#"):
            continue
        try:
            (length, filename) = re.split(r"\s+", line.strip(), maxsplit=2)
            rtr.append((filename, urlprefix+filename, int(length)))
        except Exception as e:
            raise ValueError(f"Could not parse {url} line {linenum} '{line}'") from e
    return rtr
    
def _datarepo_index_default_none(self, component, variant=None, version='latest', **kwargs):
    """
    For backward compatability with vaswani - use default for variant 
    """
    if variant is None:
        variant = 'terrier_stemmed'
    return _datarepo_index(self, component, variant=variant, version=version, **kwargs)

ANTIQUE_FILES = {
    "topics" : {
        "train" : ("antique-train-queries.txt", "http://ciir.cs.umass.edu/downloads/Antique/antique-train-queries.txt", "singleline"),
        "test" : ("antique-test-queries.txt", "http://ciir.cs.umass.edu/downloads/Antique/antique-test-queries.txt", "singleline"),
    },
    "qrels" : {
        "train" : ("antique-train.qrel", "http://ciir.cs.umass.edu/downloads/Antique/antique-train.qrel", "singleline"),
        "test" : ("antique-test.qrel", "http://ciir.cs.umass.edu/downloads/Antique/antique-test.qrel", "singleline"),
    },
    "corpus" : 
        [("antique-collection.txt", "http://ciir.cs.umass.edu/downloads/Antique/antique-collection.txt")],
    "info_url" : "https://ciir.cs.umass.edu/downloads/Antique/readme.txt",
    "corpus_iter" : passage_generate
}

TREC_COVID_FILES = {
    "topics" : {
        "round1" : ("topics-rnd1.xml", "https://ir.nist.gov/covidSubmit/data/topics-rnd1.xml", "trecxml"),
        "round2" : ("topics-rnd2.xml", "https://ir.nist.gov/covidSubmit/data/topics-rnd2.xml", "trecxml"),
        "round3" : ("topics-rnd3.xml", "https://ir.nist.gov/covidSubmit/data/topics-rnd3.xml", "trecxml"),
        "round4" : ("topics-rnd4.xml", "https://ir.nist.gov/covidSubmit/data/topics-rnd4.xml", "trecxml"),
        "round5" : ("topics-rnd5.xml", "https://ir.nist.gov/covidSubmit/data/topics-rnd5.xml", "trecxml"),
    },
    "qrels" : {
        "round1" : ("qrels-rnd1.txt", "https://ir.nist.gov/covidSubmit/data/qrels-rnd1.txt"),
        "round2" : ("qrels-rnd2.txt", "https://ir.nist.gov/covidSubmit/data/qrels-rnd2.txt"),
        "round3" : ("qrels-rnd3.txt", "https://ir.nist.gov/covidSubmit/data/qrels-covid_d3_j2.5-3.txt"),
        "round3-cumulative" : ("qrels-rnd3-cumulative.txt", "https://ir.nist.gov/covidSubmit/data/qrels-covid_d3_j0.5-3.txt"),
        "round4" : ("qrels-rnd4.txt", "https://ir.nist.gov/covidSubmit/data/qrels-covid_d4_j3.5-4.txt"),
        "round4-cumulative" : ("qrels-rnd4-cumulative.txt", "https://ir.nist.gov/covidSubmit/data/qrels-covid_d4_j0.5-4.txt"),
        "round5" : ("qrels-covid_d5_j4.5-5.txt", "https://ir.nist.gov/covidSubmit/data/qrels-covid_d5_j4.5-5.txt"),
    },
    "corpus" : {
        "round4": ("round4.tar.gz", "https://ai2-semanticscholar-cord-19.s3-us-west-2.amazonaws.com/historical_releases/cord-19_2020-06-19.tar.gz"),
        "round5": ("round5.tar.gz", "https://ai2-semanticscholar-cord-19.s3-us-west-2.amazonaws.com/historical_releases/cord-19_2020-07-16.tar.gz"), 
    },
    "docids" : { 
        "docids-rnd3" : ("docids-rnd3.txt", "https://ir.nist.gov/covidSubmit/data/docids-rnd3.txt"),
        "docids-rnd4" : ("docids-rnd4.txt", "https://ir.nist.gov/covidSubmit/data/docids-rnd4.txt"),
        "docids-rnd5" : ("docids-rnd5.txt", "https://ir.nist.gov/covidSubmit/data/docids-rnd5.txt")
    },
    "info_url"  : "https://ir.nist.gov/covidSubmit/",
    "index": _datarepo_index
}

def msmarco_document_generate(dataset):
    for filename in dataset.get_corpus(variant="corpus-tsv"):
        with pt.io.autoopen(filename, 'rt') as corpusfile:
            for line in corpusfile: #for each line
                docno, url, title, passage = line.split("\t")
                yield {'docno' : docno, 'url' : url, 'title' : title, 'text' : passage}

MSMARCO_DOC_FILES = {
    "corpus" : 
        [("msmarco-docs.trec.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-docs.trec.gz")],
    "corpus-tsv":
        [("msmarco-docs.tsv.gz",  "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-docs.tsv.gz")],
    "topics" : 
        { 
            "train" : ("msmarco-doctrain-queries.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-doctrain-queries.tsv.gz", "singleline"),
            "dev" : ("msmarco-docdev-queries.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-docdev-queries.tsv.gz", "singleline"),
            "test" : ("msmarco-test2019-queries.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-test2019-queries.tsv.gz", "singleline"),
            "test-2020" : ("msmarco-test2020-queries.tsv.gz" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-test2020-queries.tsv.gz", "singleline"),
            'leaderboard-2020' : ("docleaderboard-queries.tsv.gz" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/docleaderboard-queries.tsv.gz", "singleline")
        },
    "qrels" : 
        { 
            "train" : ("msmarco-doctrain-qrels.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-doctrain-qrels.tsv.gz"),
            "dev" : ("msmarco-docdev-qrels.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-docdev-qrels.tsv.gz"),
            "test" : ("2019qrels-docs.txt", "https://trec.nist.gov/data/deep/2019qrels-docs.txt"),
            "test-2020" : ("2020qrels-docs.txt", "https://trec.nist.gov/data/deep/2020qrels-docs.txt")
        },
    "info_url" : "https://microsoft.github.io/msmarco/",
    "corpus_iter" : msmarco_document_generate,
    "index" : _datarepo_index
}

MSMARCO_PASSAGE_FILES = {
    "corpus" : 
        [("collection.tsv", "collection.tar.gz#collection.tsv")],
    "topics" :
        { 
            "train" : ("queries.train.tsv", "queries.tar.gz#queries.train.tsv", "singleline"),
            "dev" : ("queries.dev.tsv", "queries.tar.gz#queries.dev.tsv", "singleline"),
            "dev.small" : ("queries.dev.small.tsv", "collectionandqueries.tar.gz#queries.dev.small.tsv", "singleline"),
            "eval" : ("queries.eval.tsv", "queries.tar.gz#queries.eval.tsv", "singleline"),
            "eval.small" : ("queries.eval.small.tsv", "collectionandqueries.tar.gz#queries.eval.small.tsv", "singleline"),
            "test-2019" : ("msmarco-test2019-queries.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-test2019-queries.tsv.gz", "singleline"),
            "test-2020" : ("msmarco-test2020-queries.tsv.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-test2020-queries.tsv.gz", "singleline")
        },        
    "tars" : {
        "queries.tar.gz" : ("queries.tar.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/queries.tar.gz"),
        "collection.tar.gz" : ("collection.tar.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/collection.tar.gz"),
        "collectionandqueries.tar.gz" : ("collectionandqueries.tar.gz", "https://msmarco.z22.web.core.windows.net/msmarcoranking/collectionandqueries.tar.gz")
    },
    "qrels" : 
        { 
            "train" : ("qrels.train.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/qrels.train.tsv"),
            "dev" : ("qrels.dev.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/qrels.dev.tsv"),
            "test-2019" : ("2019qrels-docs.txt", "https://trec.nist.gov/data/deep/2019qrels-pass.txt"),
            "test-2020" : ("2020qrels-docs.txt", "https://trec.nist.gov/data/deep/2020qrels-pass.txt"),
            "dev.small" : ("qrels.dev.small.tsv", "collectionandqueries.tar.gz#qrels.dev.small.tsv"),
        },
    "info_url" : "https://microsoft.github.io/MSMARCO-Passage-Ranking/",
    "corpus_iter" : passage_generate,
    "index" : _datarepo_index
}

MSMARCOv2_DOC_FILES = {
    "info_url" : "https://microsoft.github.io/msmarco/TREC-Deep-Learning.html",
    "topics" : {
        "train" : ("docv2_train_queries.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_train_queries.tsv", "singleline"),
        "dev1"  :("docv2_dev_queries.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_dev_queries.tsv", "singleline"),
        "dev2"  :("docv2_dev2_queries.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_dev2_queries.tsv", "singleline"),
        "valid1" : ("msmarco-test2019-queries.tsv.gz" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-test2019-queries.tsv.gz", "singleline"),
        "valid2" : ("msmarco-test2020-queries.tsv.gz" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/msmarco-test2020-queries.tsv.gz", "singleline"),
        "trec_2021" : ("2021_queries.tsv" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/2021_queries.tsv", "singleline"),
    },
    "qrels" : {
        "train" : ("docv2_train_qrels.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_train_qrels.tsv"),
        "dev1"  :("docv2_dev_qrels.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_dev_qrels.tsv"),
        "dev2"  :("docv2_dev2_qrels.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_dev2_qrels.tsv"),
        "valid1" : ("docv2_trec2019_qrels.txt.gz" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_trec2019_qrels.txt.gz"),
        "valid2" : ("docv2_trec2020_qrels.txt.gz" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/docv2_trec2020_qrels.txt.gz")
    },
    "index" : _datarepo_index,
}

MSMARCOv2_PASSAGE_FILES = {
    "info_url" : "https://microsoft.github.io/msmarco/TREC-Deep-Learning.html",
    "topics" : {
        "train" : ("passv2_train_queries.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/passv2_train_queries.tsv", "singleline"),
        "dev1"  : ("passv2_dev_queries.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/passv2_dev_queries.tsv", "singleline"),
        "dev2"  : ("passv2_dev2_queries.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/passv2_dev2_queries.tsv", "singleline"),
        "trec_2021" : ("2021_queries.tsv" , "https://msmarco.z22.web.core.windows.net/msmarcoranking/2021_queries.tsv", "singleline"),
    },
    "qrels" : {
        "train" : ("passv2_train_qrels.tsv" "https://msmarco.z22.web.core.windows.net/msmarcoranking/passv2_train_qrels.tsv"),
        "dev1"  : ("passv2_dev_qrels.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/passv2_dev_qrels.tsv"),
        "dev2"  : ("passv2_dev2_qrels.tsv", "https://msmarco.z22.web.core.windows.net/msmarcoranking/passv2_dev2_qrels.tsv"),
    },
    "index" : _datarepo_index,
}

# remove WT- prefix from topics
def remove_prefix(self, component, variant):
    topics_file, type = self._get_one_file("topics_prefixed", variant)
    if type in pt.io.SUPPORTED_TOPICS_FORMATS:
        topics = pt.io.read_topics(topics_file, type)
    else:
        raise ValueError("Unknown topic type %s" % type)
    topics["qid"] = topics.apply(lambda row: row["qid"].split("-")[1], axis=1)
    return (topics, "direct")


# a function to fix the namedpage TREC Web tracks 2001 and 2002
def parse_desc_only(self, component, variant):
    file, type = self._get_one_file("topics_desc_only", variant=variant)
    topics = pt.io.read_topics(file, format="trec", whitelist=["DESC"], blacklist=None)
    topics["qid"] = topics.apply(lambda row: row["qid"].replace("NP", ""), axis=1)
    topics["qid"] = topics.apply(lambda row: row["qid"].replace("EP", ""), axis=1)
    return (topics, "direct")

TREC_WT_2002_FILES = {
    "topics" : 
        { 
            "td" : ("webtopics_551-600.txt.gz", "https://trec.nist.gov/data/topics_eng/webtopics_551-600.txt.gz", "trec"),
            "np" : parse_desc_only
        },
    "topics_desc_only" : {
        "np" : ("webnamed_page_topics.1-150.txt.gz", "https://trec.nist.gov/data/topics_eng/webnamed_page_topics.1-150.txt.gz", "trec")
    },
    "qrels" : 
        { 
            "np" : ("qrels.named-page.txt.gz", "https://trec.nist.gov/data/qrels_eng/qrels.named-page.txt.gz"),
            "td" : ("qrels.distillation.txt.gz", "https://trec.nist.gov/data/qrels_eng/qrels.distillation.txt.gz")
        },
    "info_url" : "https://trec.nist.gov/data/t11.web.html",
}

TREC_WT_2003_FILES = {
    "topics" : 
        { 
            "np" : ("webtopics_551-600.txt.gz", "https://trec.nist.gov/data/topics_eng/webtopics_551-600.txt.gz", "trec"),
            "td" : ("2003.distillation_topics.1-50.txt", "https://trec.nist.gov/data/topics_eng/2003.distillation_topics.1-50.txt", "trec"),
        },
    "qrels" : 
        { 
            "np" : ("qrels.named-page.txt.gz", "https://trec.nist.gov/data/qrels_eng/qrels.named-page.txt.gz"),
            "td" : ("qrels.distillation.2003.txt", "https://trec.nist.gov/data/qrels_eng/qrels.distillation.2003.txt")
        },
    "info_url" : "https://trec.nist.gov/data/t12.web.html",
}

def irds_mirror(md5):
    return f'http://mirror.ir-datasets.com/{md5}'

def filter_on_qid_type(self, component, variant):
    if component == "topics":
        data = self.get_topics("all")
    elif component == "qrels":
        data = self.get_qrels("all")
    qid2type_file = self._get_one_file("topics_map")[0]
    qid2type = pd.read_csv(qid2type_file, names=["qid", "type"], sep=" ")
    qid2type["qid"] = qid2type.apply(lambda row: row["qid"].split("-")[1], axis=1)
    rtr = data.merge(qid2type[qid2type["type"] == variant], on=["qid"])
    if len(rtr) == 0:
        raise ValueError("No such topic type '%s'" % variant)
    rtr.drop(columns=['type'], inplace=True)
    return (rtr, "direct")

TREC_WT_2004_FILES = {
    "topics" : 
        { 
            "all" : remove_prefix,
            "np": filter_on_qid_type,
            "hp": filter_on_qid_type,
            "td": filter_on_qid_type,
        },
    "topics_map" : [("04.topic-map.official.txt", [
        "https://trec.nist.gov/data/web/04.topic-map.official.txt",
        irds_mirror("79737768b3be1aa07b14691aa54802c5"),
        "https://www.dcs.gla.ac.uk/~craigm/04.topic-map.official.txt"
        ] )],
    "topics_prefixed" : { 
        "all" : ("Web2004.query.stream.trecformat.txt", [
                "https://trec.nist.gov/data/web/Web2004.query.stream.trecformat.txt",
                irds_mirror("10821f7a000b8bec058097ede39570be"),
                "https://www.dcs.gla.ac.uk/~craigm/Web2004.query.stream.trecformat.txt"], 
            "trec")
    },
    "qrels" : 
        {
            "hp" : filter_on_qid_type,
            "td" : filter_on_qid_type,
            "np" : filter_on_qid_type,
            "all" : ("04.qrels.web.mixed.txt", [
                "https://trec.nist.gov/data/web/04.qrels.web.mixed.txt",
                irds_mirror("93daa0e4b4190c84e30d2cce78a0f674"),
                "https://www.dcs.gla.ac.uk/~craigm/04.qrels.web.mixed.txt"])
        },
    "info_url" : "https://trec.nist.gov/data/t13.web.html",
}


# a function for the TREC Web track 2009 qrels, to make prels into qrels
def prel2qrel(self, component, variant): 
    prel_file, _ = self._get_one_file("prels", variant)
    df = pd.read_csv(prel_file, sep=" ", names=["qid", "docno", "label", "oth1", "oth2"])[["qid", "docno", "label"]]
    df["qid"] = df["qid"].astype(str)
    df["docno"] = df["docno"].astype(str)
    return (df, "direct")

TREC_WT_2009_FILES = {
    "topics" : [  
            remove_prefix
        ],

    "topics_prefixed" : [  
            ("wt09.topics.queries-only", "https://trec.nist.gov/data/web/09/wt09.topics.queries-only", "singleline")
        ],
    "qrels" :  {
        "adhoc" : prel2qrel, 
        "adhoc.catA" : prel2qrel,
        "adhoc.catB" : prel2qrel,
    },
    "prels" : {
        "adhoc" : ("prels.1-50.gz", "https://trec.nist.gov/data/web/09/prels.1-50.gz"),
        "adhoc.catA" : ("prels.catA.1-50.gz", "https://trec.nist.gov/data/web/09/prels.catA.1-50.gz"),
        "adhoc.catB" : ("prels.catB.1-50.gz", "https://trec.nist.gov/data/web/09/prels.catB.1-50.gz")
    },
    "info_url" : "https://trec.nist.gov/data/web09.html",
}

TREC_WT_2010_FILES = {
     "topics" : [  
            ("wt2010-topics.queries-only", "https://trec.nist.gov/data/web/10/wt2010-topics.queries-only", "singleline")
        ],
    "qrels" : 
        { 
            "adhoc" : ("qrels.adhoc", "https://trec.nist.gov/data/web/10/10.adhoc-qrels.final")
        },
    "info_url" : "https://trec.nist.gov/data/web10.html",
}

TREC_WT_2011_FILES = {
     "topics" : [  
            ("queries.101-150.txt", "https://trec.nist.gov/data/web/11/queries.101-150.txt", "singleline")
        ],
    "qrels" : 
        { 
            "adhoc" : ("qrels.adhoc", "https://trec.nist.gov/data/web/11/qrels.adhoc")
        },
    "info_url" : "https://trec.nist.gov/data/web2011.html",
}

TREC_WT_2012_FILES = {
     "topics" : [  
            ("queries.151-200.txt", "https://trec.nist.gov/data/web/12/queries.151-200.txt", "singleline")
        ],
    "qrels" : 
        { 
            "adhoc" : ("qrels.adhoc", "https://trec.nist.gov/data/web/12/qrels.adhoc")
        },
    "info_url" : "https://trec.nist.gov/data/web2012.html",
}

TREC_WT2G_FILES = {
    "qrels" : [ ("qrels.trec8.small_web.gz", "https://trec.nist.gov/data/qrels_eng/qrels.trec8.small_web.gz") ],
    "topics" : [ (  "topics.401-450.gz", "https://trec.nist.gov/data/topics_eng/topics.401-450.gz" ) ],
    "info_url" : "https://trec.nist.gov/data/t8.web.html",
}

TREC_WT10G_FILES = {
    "qrels" : {
        "trec9" : ("qrels.trec9.main_web.gz", "https://trec.nist.gov/data/qrels_eng/qrels.trec9.main_web.gz"),
        "trec10-adhoc" : ("qrels.trec10.main_web.gz", "https://trec.nist.gov/data/qrels_eng/qrels.trec10.main_web.gz"),
        "trec10-hp" : ("qrels.trec10.entrypage.gz", "https://trec.nist.gov/data/qrels_eng/qrels.trec10.entrypage.gz"),
    },
    "topics" : {
        "trec9" : (  "topics.451-500.gz", "https://trec.nist.gov/data/topics_eng/topics.451-500.gz" ),
        "trec10-adhoc" : (  "topics.501-550.txt", "https://trec.nist.gov/data/topics_eng/topics.501-550.txt" ),
         "trec10-hp" : parse_desc_only
    },
    "topics_desc_only" : {
         "trec10-hp" : (  "entry_page_topics.1-145.txt", "https://trec.nist.gov/data/topics_eng/entry_page_topics.1-145.txt" ),
    },
    "info_url" : "https://trec.nist.gov/data/t9.web.html",
}

def _merge_years(self, component, variant):
    MAP_METHOD = { 
        "topics" : pt.datasets.RemoteDataset.get_topics,
        "qrels" : pt.datasets.RemoteDataset.get_qrels,  
    }
    dfs = []
    low, hi = variant.split("-")
    for y in range(int(low), int(hi)+1):
        df = MAP_METHOD[component](self, variant=str(y))
        dfs.append(df)
    return (pd.concat(dfs), "direct")

TREC_TB_FILES = {
    "topics" : {
        "2004" : ( "04topics.701-750.txt", "https://trec.nist.gov/data/terabyte/04/04topics.701-750.txt" ),
        "2005" : ( "04topics.701-750.txt", "https://trec.nist.gov/data/terabyte/05/05.topics.751-800.txt" ),
        "2006" : ( "06.topics.801-850.txt", "https://trec.nist.gov/data/terabyte/06/06.topics.801-850.txt" ),
        "2004-2006" : ("06.topics.701-850.txt", "https://trec.nist.gov/data/terabyte/06/06.topics.701-850.txt"),

        "2006-np" : ( "06.np_topics.901-1081.txt", "https://trec.nist.gov/data/terabyte/06/06.np_topics.901-1081.txt" ),
        "2005-np" : ( "05.np_topics.601-872.final.txt", "https://trec.nist.gov/data/terabyte/05/05.np_topics.601-872.final.txt")
    },

    "qrels" : {
        "2004" : ( "04.qrels.12-Nov-04", "https://trec.nist.gov/data/terabyte/04/04.qrels.12-Nov-04"),
        "2005" : ( "05.adhoc_qrels", "https://trec.nist.gov/data/terabyte/05/05.adhoc_qrels"),
        "2006" : ( "qrels.tb06.top50", "https://trec.nist.gov/data/terabyte/06/qrels.tb06.top50"),
        "2004-2006" : _merge_years,

        "2005-np" : ( "05.np_qrels", "https://trec.nist.gov/data/terabyte/05/05.np_qrels"),
        "2006-np" : ( "qrels.tb06.np", "https://trec.nist.gov/data/terabyte/06/qrels.tb06.np"),
    },
    "info_url" : "https://trec.nist.gov/data/terabyte.html"
}

TREC_ROBUST_04_FILES = {
    "qrels" : [ ("qrels.robust2004.txt", "https://trec.nist.gov/data/robust/qrels.robust2004.txt") ],
    "topics" : [ (  "04.testset.gz", "https://trec.nist.gov/data/robust/04.testset.gz" ) ],
    "info_url" : "https://trec.nist.gov/data/t13_robust.html",
}
TREC_ROBUST_05_FILES = {
    "qrels" : [ ("TREC2005.qrels.txt", "https://trec.nist.gov/data/robust/05/TREC2005.qrels.txt") ],
    "topics" : [ (  "05.50.topics.txt", "https://trec.nist.gov/data/robust/05/05.50.topics.txt" ) ],
    "info_url" : "https://trec.nist.gov/data/t14_robust.html",
}

TREC_PRECISION_MEDICINE_FILES = {
    "topics" : {
        "2017" : ("topics2017.xml", "http://www.trec-cds.org/topics2017.xml", "trecxml"),
        "2018" : ("topics2018.xml", "http://www.trec-cds.org/topics2018.xml", "trecxml"),
        "2019" : ("topics2019.xml", "http://www.trec-cds.org/topics2019.xml", "trecxml"),
        "2020" : ("topics2020.xml", "http://www.trec-cds.org/topics2020.xml", "trecxml")
    },
    "qrels" : {
        "qrels-2017-abstracts" : ("qrels-2017-abstracts.txt", "https://trec.nist.gov/data/precmed/qrels-final-abstracts.txt"),  #TODO keep original names?
        "qrels-2017-abstracts-sample" : ("qrels-2017-abstracts-sample.txt", "https://trec.nist.gov/data/precmed/sample-qrels-final-abstracts.txt"),
        "qrels-2017-trials" : ("qrels-2017-trials.txt", "https://trec.nist.gov/data/precmed/qrels-final-trials.txt"),
        "qrels-2018-abstracts" : ("qrels-2018-abstracts.txt", "https://trec.nist.gov/data/precmed/qrels-treceval-abstracts-2018-v2.txt"),
        "qrels-2018-abstracts-sample" : ("qrels-2018-abstracts-sample.txt", "https://trec.nist.gov/data/precmed/qrels-sample-abstracts-v2.txt"),
        "qrels-2018-trials" : ("qrels-2018-trials.txt", "https://trec.nist.gov/data/precmed/qrels-treceval-clinical_trials-2018-v2.txt"),
        "qrels-2018-trials-sample" : ("qrels-2018-trials-sample.txt", "https://trec.nist.gov/data/precmed/qrels-sample-trials-v2.txt"),
        "qrels-2019-abstracts" : ("qrels-2019-abstracts.txt", "https://trec.nist.gov/data/precmed/qrels-treceval-abstracts.2019.txt"),
        "qrels-2019-trials" : ("qrels-2019-trials.txt", "https://trec.nist.gov/data/precmed/qrels-treceval-trials.38.txt"),
        "qrels-2019-abstracts-sample" : ("qrels-2019-abstracts-sample.txt", "https://trec.nist.gov/data/precmed/qrels-sampleval-abstracts.2019.txt"),
        "qrels-2019-trials-sample" : ("qrels-2019-trials-sample.txt", "https://trec.nist.gov/data/precmed/qrels-sampleval-trials.38.txt")
    },
    "info_url" : "https://trec.nist.gov/data/precmed.html",
}



VASWANI_CORPUS_BASE = "https://raw.githubusercontent.com/terrier-org/pyterrier/master/tests/fixtures/vaswani_npl/"
VASWANI_INDEX_BASE = "https://raw.githubusercontent.com/terrier-org/pyterrier/master/tests/fixtures/index/"
VASWANI_FILES = {
    "corpus": [("doc-text.trec", [
        VASWANI_CORPUS_BASE + "corpus/doc-text.trec",
        irds_mirror("a059e713c50350e39999467c8c73b7c5")])],
    "topics": [("query-text.trec", [
        VASWANI_CORPUS_BASE + "query-text.trec",
        irds_mirror("3a624be2b0ef7c9534cf848891679bec")])],
    "qrels": [("qrels", [
        VASWANI_CORPUS_BASE + "qrels",
        irds_mirror("6acb6db9969da8b8c6c23c09551af8d9")])],
    "index": _datarepo_index_default_none,
    #"index":
    #    [(filename, VASWANI_INDEX_BASE + filename) for filename in STANDARD_TERRIER_INDEX_FILES + ["data.meta-0.fsomapfile"]],
    "info_url" : "http://ir.dcs.gla.ac.uk/resources/test_collections/npl/",
    "corpus_iter" : lambda dataset, **kwargs : pt.index.treccollection2textgen(dataset.get_corpus(), num_docs=11429, verbose=kwargs.get("verbose", False))
}

DATASET_MAP : Dict[str, pt.datasets.Dataset] = {
    # umass antique corpus - see http://ciir.cs.umass.edu/downloads/Antique/ 
    "antique" : pt.datasets.RemoteDataset("antique", ANTIQUE_FILES),
    # generated from http://ir.dcs.gla.ac.uk/resources/test_collections/npl/
    "vaswani": pt.datasets.RemoteDataset("vaswani", VASWANI_FILES),
    "msmarco_document" : pt.datasets.RemoteDataset("msmarco_document", MSMARCO_DOC_FILES),
    "msmarcov2_document" : pt.datasets.RemoteDataset("msmarcov2_document", MSMARCOv2_DOC_FILES),
    "msmarco_passage" : pt.datasets.RemoteDataset("msmarco_passage", MSMARCO_PASSAGE_FILES),
    "msmarcov2_passage" : pt.datasets.RemoteDataset("msmarcov2_passage", MSMARCOv2_PASSAGE_FILES),
    "trec-robust-2004" : pt.datasets.RemoteDataset("trec-robust-2004", TREC_ROBUST_04_FILES),
    "trec-robust-2005" : pt.datasets.RemoteDataset("trec-robust-2005", TREC_ROBUST_05_FILES),
    "trec-terabyte" : pt.datasets.RemoteDataset("trec-terabyte", TREC_TB_FILES),
    #medical-like tracks
    "trec-precision-medicine" : pt.datasets.RemoteDataset("trec-precicion-medicine", TREC_PRECISION_MEDICINE_FILES),
    "trec-covid" : pt.datasets.RemoteDataset("trec-covid", TREC_COVID_FILES),
    #wt2g
    "trec-wt2g" : pt.datasets.RemoteDataset("trec-wt2g", TREC_WT2G_FILES),
    #wt10g
    "trec-wt10g" : pt.datasets.RemoteDataset("trec-wt10g", TREC_WT10G_FILES),
    #.gov
    "trec-wt-2002" : pt.datasets.RemoteDataset("trec-wt-2002", TREC_WT_2002_FILES),
    "trec-wt-2003" : pt.datasets.RemoteDataset("trec-wt-2003", TREC_WT_2002_FILES),
    "trec-wt-2004" : pt.datasets.RemoteDataset("trec-wt-2004", TREC_WT_2004_FILES),
    #clueweb09
    "trec-wt-2009" : pt.datasets.RemoteDataset("trec-wt-2009", TREC_WT_2009_FILES),
    "trec-wt-2010" : pt.datasets.RemoteDataset("trec-wt-2010", TREC_WT_2010_FILES),
    "trec-wt-2011" : pt.datasets.RemoteDataset("trec-wt-2011", TREC_WT_2011_FILES),
    "trec-wt-2012" : pt.datasets.RemoteDataset("trec-wt-2012", TREC_WT_2012_FILES),
}

# "trec-deep-learning-docs"
#DATASET_MAP['msmarco_document'] = DATASET_MAP["trec-deep-learning-docs"]
#DATASET_MAP['msmarco_passage'] = DATASET_MAP["trec-deep-learning-passages"]
DATASET_MAP["trec-deep-learning-docs"] = DATASET_MAP['msmarco_document']
DATASET_MAP["trec-deep-learning-passages"] = DATASET_MAP['msmarco_passage']


class BuiltinDatasetProvider(pt.datasets.DatasetProvider):
    def get_dataset(self, name: str) -> pt.datasets.Dataset:
        return DATASET_MAP[name]

    def list_dataset_names(self) -> Iterable[str]:
        return list(DATASET_MAP.keys())


import os
from setuptools import setup

def get_version(rel_path):
    suffix = os.environ["PYTERRIER_VERSION_SUFFIX" ] if "PYTERRIER_VERSION_SUFFIX" in os.environ else ""
    for line in open(rel_path):
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1] + suffix
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name="pyterrier-flake8-ext",
    version=get_version("pyterrier_flake8_ext/__init__.py"),
    author="Sean MacAvaney",
    author_email='sean.macavaney@glasgow.ac.uk',
    description="Terrier IR platform Python API",
    project_urls={
        'Documentation': 'https://pyterrier.readthedocs.io',
        'Changelog': 'https://github.com/terrier-org/pyterrier/releases',
        'Issue Tracker': 'https://github.com/terrier-org/pyterrier/issues',
        'CI': 'https://github.com/terrier-org/pyterrier/actions',
    },
    long_description=open('README.md', 'rt').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/terrier-org/pyterrier",
    packages=['pyterrier_flake8_ext'],
    entry_points={
        'flake8.extension': [
            'PT100 = pyterrier_flake8_ext:JavaCheck',
        ]
    },
    install_requires=[
        'flake8',
    ],
    python_requires='>=3.8',
)



from packaging.version import Version
from pyterrier.java import mavenresolver
import unittest
import shutil
import tempfile
from os import path

class TestMaven(unittest.TestCase):
    ''' This doesn't rely on Terrier, so doesn't inherit from BaseTestCase '''

    def __init__(self, *args, **kwargs):
        super(TestMaven, self).__init__(*args, **kwargs)

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        try:
            shutil.rmtree(self.test_dir)
        except:
            pass

    def testVersion(self):
        ver = mavenresolver.latest_version_num("org.terrier", "terrier-core")
        self.assertGreaterEqual(Version(ver), Version('5.3'))

    def testDownload(self):
        ver = mavenresolver.latest_version_num("org.terrier", "terrier-python-helper")
        jar = mavenresolver.get_package_jar("org.terrier", "terrier-python-helper", ver, self.test_dir)
        self.assertTrue(path.exists(jar))

    def testJitpack(self):
        jar = mavenresolver.get_package_jar("com.github.terrierteam", "terrier-ciff", "-SNAPSHOT", self.test_dir)
        self.assertTrue(path.exists(jar))

if __name__ == "__main__":
    unittest.main()



from .base import TempDirTestCase
import pandas as pd
import pickle
import pyterrier as pt
import tempfile, shutil
from os import path
import os

class TestPickle(TempDirTestCase):

    def _fix_joblib(self):
        import joblib
        def dumps(obj):
            filename = path.join(self.test_dir, 'obj')
            f = open(filename, 'wb')
            joblib.dump(obj, f)
            f.close()
            f = open(filename, mode='rb')
            all_data = f.read()
            f.close()
            os.remove(filename)
            return all_data
        def loads(data):
            filename = path.join(self.test_dir, 'obj')
            f = open(filename, 'wb')
            f.write(data)
            f.close()
            obj = joblib.load(filename)
            os.remove(filename)
            return obj
        
        joblib.__dict__['dumps'] = dumps
        joblib.__dict__['loads'] = loads

    def _indexref(self, pickler):
        vaswani = pt.datasets.get_dataset("vaswani")
        ref = vaswani.get_index()
        b_ref = pickler.dumps(ref)
        ref2 = pickler.loads(b_ref)
        index = pt.IndexFactory.of(ref2)
        self.assertTrue(index.getCollectionStatistics().getNumberOfDocuments() > 0)
        index.close()

    def test_indexref_pickle(self):
        self._indexref(pickle)

    def test_indexref_joblib(self):
        import joblib
        self._fix_joblib()
        self._indexref(joblib)

    def _sourcetransformer(self, pickler):
        df = pd.DataFrame([["q1", "doc1", 5]], columns=["qid", "docno", "score"])
        t = pt.Transformer.from_df(df)
        t_2 = pickler.loads(pickler.dumps(t))
        q  = pd.DataFrame([["q1", "query"]], columns=["qid", "query"])
        res = t_2(q)
        self.assertEqual(1, len(res))
        self.assertEqual("q1", res.iloc[0]["qid"])

    def test_sourcetransformer_pickle(self):
        self._sourcetransformer(pickle)

    def test_sourcetransformer_joblib(self):
        import joblib
        self._fix_joblib()
        self._sourcetransformer(joblib)

    def test_br_pickle(self):
        self._br(pickle)

    # def test_br_dill_callback(self):
    #     import dill
    #     self._br(dill, wmodel=lambda keyFreq, posting, entryStats, collStats: posting.getFrequency())

    def test_br_pickle_callback(self):
        import pickle
        self._br(pickle, wmodel=lambda keyFreq, posting, entryStats, collStats: posting.getFrequency())

    def test_br_joblib_callback(self):
        import joblib
        self._fix_joblib()
        self._br(joblib, wmodel=lambda keyFreq, posting, entryStats, collStats: posting.getFrequency())

    def test_br_pickle_straightwmodel(self):
        self._br(pickle, wmodel=pt.java.autoclass("org.terrier.matching.models.BM25")())

    def test_br_joblib(self):
        import joblib
        self._fix_joblib()
        self._br(joblib)

    def test_fbr_pickle(self):
        self._fbr(pickle)

    def test_fbr_joblib(self):
        import joblib
        self._fix_joblib()
        self._fbr(joblib)

    def test_qe_pickle(self):
        self._qe(pickle)

    def _qe(self, pickler):
        vaswani = pt.datasets.get_dataset("vaswani")
        index = vaswani.get_index()
        bm25 = pt.terrier.Retriever(index, wmodel='BM25', controls={"c" : 0.75}, num_results=15)
        br = bm25 >> pt.rewrite.Bo1QueryExpansion(index) >> bm25
        q  = pd.DataFrame([["q1", "chemical"]], columns=["qid", "query"])
        res1 = br(q)
        byterep = pickler.dumps(br)
        br2 = pickler.loads(byterep)

        res2 = br2(q)
        pd.testing.assert_frame_equal(res1, res2)
    
    def _br(self, pickler, wmodel='BM25'):
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.Retriever(vaswani.get_index(), wmodel=wmodel, controls={"c" : 0.75}, num_results=15)
        q  = pd.DataFrame([["q1", "chemical"]], columns=["qid", "query"])
        res1 = br(q)
        byterep = pickler.dumps(br)
        br2 = pickler.loads(byterep)

        if isinstance(wmodel, str):
            self.assertEqual(wmodel, br2.controls["wmodel"])
        self.assertEqual(br.controls, br2.controls)
        self.assertEqual(br.properties, br2.properties)
        self.assertEqual(br.metadata, br2.metadata)

        self.assertEqual(repr(br), repr(br2))
        res2 = br2(q)
        
        pd.testing.assert_frame_equal(res1, res2)

    def _fbr(self, pickler):
        vaswani = pt.datasets.get_dataset("vaswani")
        br = pt.terrier.FeaturesRetriever(vaswani.get_index(), wmodel="BM25", features=["WMODEL:DPH"], controls={"c" : 0.75}, num_results=15)
        q  = pd.DataFrame([["q1", "chemical"]], columns=["qid", "query"])
        res1 = br(q)
        byterep = pickler.dumps(br)
        br2 = pickler.loads(byterep)

        self.assertEqual("BM25", br2.controls["wmodel"])
        self.assertEqual(br.controls, br2.controls)
        self.assertEqual(br.properties, br2.properties)
        self.assertEqual(br.metadata, br2.metadata)
        self.assertEqual(br.features, br2.features)

        self.assertEqual(repr(br), repr(br2))
        res2 = br2(q)
        
        pd.testing.assert_frame_equal(res1, res2)
        



from typing import List


def terrier_artifact_metadata_adapter(path: str, dir_listing: List[str]):
    """Guess whether this path is a terrier index.

    Some terrier indexes are missing pt_meta.json, but we can assume they are terrier indexes based on the
    presence of a data.properties file.
    """
    if 'data.properties' in dir_listing:
        return {
            'type': 'sparse_index',
            'format': 'terrier',
            'package_hint': 'python-terrier',
        }



from typing import Optional
import os
import json
import requests
import bibtexparser
import gzip
from pylatexenc.latex2text import LatexNodes2Text
from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx.domains.std import StandardDomain

latex_converter = LatexNodes2Text()

# Cache file to store BibTeX entries
CACHE_FILE = "dblp_cache.json.gz"


def load_dblp_cache():
    """Load the cache from the file, or create a new one if it doesn't exist."""
    if os.path.exists(CACHE_FILE):
        with gzip.open(CACHE_FILE, "rt") as f:
            try:
                return json.load(f)
            except json.JSONDecodeError:
                return {}
    return {}

dblp_cache = load_dblp_cache()

def save_dblp_cache():
    """Save the cache to the file."""
    with gzip.open(CACHE_FILE, "wt") as f:
        json.dump(dblp_cache, f, indent=4)


class CiteNode(nodes.General, nodes.Element):
    """Custom node to render DBLP citation."""
    def __init__(self, citation_id, citation, link, bibtex):
        super().__init__()
        self['id'] = citation_id
        self['citation'] = citation
        self['link'] = link
        self['bibtex'] = bibtex


class CiteDirective(Directive):
    required_arguments = 1
    has_content = True
    option_spec = {'citation': str, 'link': str, 'blbtex': str}

    def run(self):
        # Create a custom node to hold the BibTeX content and display data
        node = CiteNode(
            self.get_citation_id(),
            self.get_citation(),
            self.get_link(),
            self.get_bibtex()
        )

        # in the future, could store citation information in self.state.document.settings.env
        # to generate a bibliography page.
        self.state.document.settings.env.bibliography.setdefault(node['id'], {
            'id': node['id'],
            'citation': node['citation'],
            'link': node['link'],
            'bibtex': node['bibtex'],
            'refs': {},
        })['refs'][self.state.document.settings.env.docname] = {
            'target': self.state.document.settings.env.docname,
        }

        return [node]

    def get_citation_id(self) -> str:
        return self.arguments[0]

    def get_citation(self) -> Optional[str]:
        return self.options.get('citation')

    def get_link(self) -> Optional[str]:
        return self.options.get('link')

    def get_bibtex(self) -> Optional[str]:
        if self.content:
            return '\n'.join(self.content)
        return None


class CiteDblpDirective(CiteDirective):
    """Directive to fetch and display a citation from DBLP using a DBLP ID."""
    has_content = False
    required_arguments = 1  # The DBLP ID must be provided as an argument.

    def run(self):
        dblp_id = self.arguments[0]

        if dblp_id in dblp_cache:
            bibtex_entry_short, bibtex_entry_full = dblp_cache[dblp_id]
        else:
            # Fetch BibTeX entry from DBLP
            try:
                response = requests.get(f"https://dblp.org/rec/{dblp_id}.bib?param=0")
                response.raise_for_status()
                bibtex_entry_short = response.text
                response = requests.get(f"https://dblp.org/rec/{dblp_id}.bib?param=1")
                response.raise_for_status()
                bibtex_entry_full = response.text
                dblp_cache[dblp_id] = [bibtex_entry_short, bibtex_entry_full]  # Save to cache
                save_dblp_cache()  # Persist the cache
            except requests.RequestException as e:
                error_msg = f"Failed to fetch BibTeX for DBLP ID '{dblp_id}': {str(e)}"
                return [nodes.error(None, nodes.Text(error_msg))]
        self.bibtex_entry_short = bibtex_entry_short
        self.bibtex_entry_full = bibtex_entry_full
        return super().run()

    def get_citation_id(self) -> str:
        return 'dblp:' + self.arguments[0]

    def get_citation(self) -> Optional[str]:
        citation = super().get_citation()
        if citation is not None:
            return citation
        res = bibtexparser.loads(self.bibtex_entry_short).entries[0]
        authors = latex_converter.latex_to_text(res.get('author', ''))
        if authors:
            authors = authors.split(' and\n')
            if len(authors) > 2:
                authors = authors[0].split()[-1] + ' et al'
            elif len(authors) == 2:
                authors = authors[0].split()[-1] + ' and ' + authors[1].split()[-1]
            else:
                authors = authors[0].split()[-1]
        title = latex_converter.latex_to_text(res.get('title', ''))
        year = latex_converter.latex_to_text(res.get('year', ''))
        if 'journal' in res and res['journal'] == 'CoRR':
            citation = f'{authors}. {title}. arXiv {year}.'
        elif 'journal' in res:
            journal = latex_converter.latex_to_text(res['journal'])
            citation = f'{authors}. {title}. {journal} {year}.'
        elif 'booktitle' in res:
            conf = latex_converter.latex_to_text(res['booktitle'])
            citation = f'{authors}. {title}. {conf} {year}.'
        else:
            citation = f'{authors}. {title}. {year}.'
        return citation

    def get_link(self) -> Optional[str]:
        link = super().get_link()
        if link is not None:
            return link
        res = bibtexparser.loads(self.bibtex_entry_full).entries[0]
        return latex_converter.latex_to_text(res.get('url', ''))

    def get_bibtex(self) -> Optional[str]:
        bibtex = super().get_bibtex()
        if bibtex is not None:
            return bibtex
        return self.bibtex_entry_full


def get_hierarchical_title(docname, app):
    """
    Get a hierarchical title for a Sphinx document, including its position in the toctree.

    Args:
        docname (str): The docname (e.g., 'something/text').
        app (sphinx.application.Sphinx): The Sphinx application object.

    Returns:
        str: The hierarchical title, with sections separated by " > ".
    """
    def get_doc_title(docname):
        """Get the title for a given docname."""
        title_node = app.env.titles.get(docname)
        return title_node.astext() if title_node else "Untitled"

    def find_parents(docname, toctree):
        """Find the parent hierarchy of a document in the toctree."""
        for node in toctree.traverse():
            if node.tagname == "reference" and node.get("refuri") == app.builder.get_target_uri(docname):
                # Found the current doc in the toctree
                parent_node = node.parent.parent
                if parent_node.tagname == "toctree":
                    parent_docname = parent_node.get("docname")
                    if parent_docname:
                        return find_parents(parent_docname, app.env.tocs[parent_docname]) + [parent_docname]
        return []

    # Find parents by walking up the toctree
    parent_hierarchy = []
    if docname in app.env.tocs:
        parent_hierarchy = find_parents(docname, app.env.tocs[docname])

    # Collect titles for the hierarchy
    hierarchical_titles = [get_doc_title(parent) for parent in parent_hierarchy]
    hierarchical_titles.append(get_doc_title(docname))

    # Join titles with separators
    return " » ".join(hierarchical_titles)


def visit_cite_node_html(self, node):
    """HTML visitor to render the DblpNode."""
    citation = node['citation']
    link = node['link']
    bibtex = node['bibtex']

    if link:
        link_html = f' <a href="{link}" target="_blank">[link]</a>'
    else:
        link_html = ''

    if bibtex:
        body_html = f'''
<details class="dblp-citation">
  <summary><b>{citation}</b>{link_html}</summary>
  <pre class="bibtex">{bibtex.strip()}</pre>
</details>
'''
    else:
        body_html = f'<p><b>{citation}</b>{link_html}</p>'

    self.body.append(f'''
<div class="admonition">
  <p class="admonition-title">Citation</p>
  {body_html}
</div>
    ''')


def depart_cite_node_html(self, node):
    """HTML departure for DblpNode."""
    pass


def build_breadcrumb_titles(env, toc):
    result = {}
    for node in toc.traverse():
        if node.tagname == 'toctree':
            for page_title, page in node['entries']:
                if page_title is None:
                    page_title = env.titles.get(page)
                    page_title = page_title.astext() if page_title else 'Untitled'
                page_title = page_title.strip()
                result[page] = page_title
                if page in env.tocs:
                    for subpage, subpage_title in build_breadcrumb_titles(env, env.tocs[page]).items():
                        result[subpage] = f'{page_title} » {subpage_title}'
    return result

def collect_bibliiography(app):
    if app.builder.embedded or len(app.env.bibliography) == 0:
        # Building embedded (e.g. htmlhelp or ePub) or there were no bibs found
        return []

    breadcrumb_titles = build_breadcrumb_titles(app.env, app.env.tocs['index'])

    for bib in app.env.bibliography.values():
        bib['refs'] = list(bib['refs'].values())
        for ref in bib['refs']:
            ref['url'] = app.builder.get_target_uri(ref['target'])
            ref['title'] = breadcrumb_titles.get(ref['target'], 'Untitled')

    return [(
        'bibliography',
        {'bibliography': sorted(app.env.bibliography.values(), key=lambda x: x['citation'])},
        'bibliography.html',
    )]


def setup_bibliiography_env(app, env, pages):
    app.env.bibliography = {}


def setup(app):
    app.add_node(CiteNode, html=(visit_cite_node_html, depart_cite_node_html))
    app.add_directive('cite', CiteDirective)
    app.add_directive('cite.dblp', CiteDblpDirective)
    app.connect('env-before-read-docs', setup_bibliiography_env)
    app.connect('html-collect-pages', collect_bibliiography)
    StandardDomain._virtual_doc_names['bibliography'] = ('bibliography', 'Bibliography')
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import more_itertools
from collections import defaultdict
import re
import pandas as pd
from typing import Any, List, Union, Literal, Protocol, runtime_checkable
from warnings import warn
import types
import pyterrier as pt

@runtime_checkable
class HasTextLoader(Protocol):
    def text_loader(
        self,
        fields: Union[List[str], str, Literal['*']] = '*',
        *,
        verbose: bool = False,
    ) -> pt.Transformer:
        """
        Returns a transformer that loads and populates text columns for each document in the
        provided input frame.

        :param fields: The names of the fields to load. If a list of strings, all fields are provided.
            If a single string, this single field is provided. If the special value of '*' (default,
            all available fields are provided.
        :param verbose: Show a progress bar.
        """


def get_text(
        indexlike: Union[HasTextLoader, str],
        metadata : Union[str,List[str], Literal['*']] = '*',
        by_query : bool = False,
        verbose : bool = False,
        **kwargs: Any) -> pt.Transformer:
    """
    A utility transformer for obtaining the text from the text of documents (or other document metadata) from Terrier's MetaIndex
    or an IRDSDataset docstore.

    :param indexlike: an object that provides a .text_loader() factory method, such as a Terrier index or IRDSDataset.
        If a ``str`` is provided, it will try to load a Terrier index from the provided path.
    :param metadata: The names of the fields to load. If a list of strings, all fields are provided.
        If a single string, this single field is provided. If the special value of '*' (default), all
        available fields are provided.
    :param by_query: whether the entire dataframe should be progressed at once, rather than one query at a time.
        Defaults to false, which means that all document metadata will be fetched at once.
    :param verbose: whether to print a tqdm progress bar. When by_query=True, prints progress by query. Otherwise,
        the behaviour is defined by the provided ``indexlike``.
    :param kwargs: other arguments to pass through to the text_loader.
    :return: a transformer that loads the text of documents from the provided indexlike.
    :rtype: pt.Transformer
    :raises ValueError: if indexlike does not provide a .text_loader() method.

    Example (Terrier Index)::

        index = pt.IndexFactory.of("./index/")
        pipe = ( pt.terrier.Retriever(index, wmodel="DPH")
            >> pt.text.get_text(index) # load text using a PyTerrier index
            >> pt.text.scorer(wmodel="DPH") )

    Example (IR Datasets)::

        # see https://github.com/terrierteam/pyterrier_t5
        from pyterrier_t5 import MonoT5ReRanker
        bm25 = pt.terrier.Retriever.from_dataset(pt.get_dataset('msmarcov2_passage'), wmodel='BM25')
        # load text using IR Datasets
        loader = pt.text.get_text(pt.get_dataset('irds:msmarco-passage-v2'), ['text'])
        monoT5 = bm25 >> loader >> MonoT5ReRanker()

    """
    if isinstance(indexlike, str):
        # TODO: We'll need to decide how to handle this once terrier is split from core
        # Maybe it should run Artifact.load(indexlike) instead?
        indexlike = pt.IndexFactory.of(indexlike)

    if not isinstance(indexlike, HasTextLoader):
        raise ValueError('indexlike must provide a .text_loader() method.')

    result : pt.Transformer
    result = indexlike.text_loader(metadata, verbose=verbose and not by_query, **kwargs)

    if by_query:
        result = pt.apply.by_query(result, verbose=verbose)

    return result


def scorer(*args, **kwargs) -> pt.Transformer:
    """
    This allows scoring of the documents with respect to a query, without creating an index first. 
    This is an alias to pt.TextScorer(). Internally, a Terrier memory index is created, before being
    used for scoring.

    :pararm body_attr: what dataframe input column contains the text of the document. Default is `"body"`.
    :param wmodel: name of the weighting model to use for scoring.
    :param background_index: An optional background index to use for collection statistics. If a weighting
        model such as BM25 or TF_IDF or PL2 is used without setting the background_index, the background statistics
        will be calculated from the dataframe, which is ususally not the desired behaviour.
    :param args: other arguments to pass through to the TextScorer.
    :param kwargs: other arguments to pass through to the TextScorer.
    :return: a transformer that scores the documents with respect to a query.
    :rtype: pt.Transformer   

    Example::
    
        df = pd.DataFrame(
            [
                ["q1", "chemical reactions", "d1", "professor protor poured the chemicals"],
                ["q1", "chemical reactions", "d2", "chemical brothers turned up the beats"],
            ], columns=["qid", "query", "docno", "text"])
        textscorerTf = pt.text.scorer(body_attr="text", wmodel="Tf")
        rtr = textscorerTf.transform(df)
        # rtr will have a score for each document for the query "chemical reactions" based on the provided document contents
        # both attain score 1, as, after stemming, they both contain one occurrence of the query term 'chemical'
        # ["q1", "chemical reactions", "d1", "professor protor poured the chemicals", 0, 1]
        # ["q1", "chemical reactions", "d2", "chemical brothers turned up the beats", 0, 1]

    For calculating the scores of documents using any weighting model with the concept of IDF, it is strongly advised to make use of
    an existing Terrier index for background statistics. Without a background index, IDF will be calculated based on the supplied
    dataframe (for models such as BM25, this can lead to negative scores)::

        textscorerTfIdf = pt.text.scorer(body_attr="text", wmodel="TF_IDF", background_index=index)

    """
    return pt.terrier.retriever.TextScorer(*args, **kwargs)

def sliding( text_attr='body', length=150, stride=75, join=' ', prepend_attr='title', tokenizer=None, **kwargs) -> pt.Transformer:
    r"""
    A useful transformer for splitting long documents into smaller passages within a pipeline. This applies a *sliding* window over the
    text, where each passage is the give number of tokens long. Passages can overlap, if the stride is set smaller than the length. In
    applying this transformer, docnos are altered by adding '%p' and a passage number. The original scores for each document can be recovered
    by aggregation functions, such as ``max_passage()``.

    For the puposes of obtaining passages of a given length, the tokenisation can be controlled. By default, tokenisation takes place by splitting
    on space, i.e. based on the Python regular expression ``re.compile(r'\s+')``. However, more fine-grained tokenisation can applied by passing 
    an object matching the HuggingFace Transformers `Tokenizer API <https://huggingface.co/docs/transformers/main/en/main_classes/tokenizer#transformers.PreTrainedTokenizer>`_ 
    as the `tokenizer` kwarg argument. In short, the `tokenizer` object must have a ``.tokenize(str) -> list[str]`` method and 
    ``.convert_tokens_to_string(list[str]) -> str`` for detokenisation.
    
    :param text_attr: what is the name of the dataframe attribute containing the main text of the document to be split into passages.
        Default is 'body'.
    :param length: how many tokens in each passage. Default is 150.
    :param stride: how many tokens to advance each passage by. Default is 75.
    :param join: how to join the tokens of the passage together. Default is ' '.
    :param prepend_attr: whether another document attribute, such as the title of the document, to each passage, following [Dai2019]. Defaults to 'title'.
    :param tokenizer: which model to use for tokenizing. The object must have a ``.tokenize(str) -> list[str]`` method for tokenization and ``.convert_tokens_to_string(list[str]) -> str`` for detokenization.
            Default is None. Tokenisation is perfomed by splitting on one-or-more spaces, i.e. based on the Python regular expression ``re.compile(r'\s+')``
    :param kwargs: other arguments to pass through to the SlidingWindowPassager.
    :return: a transformer that splits the documents into passages.
    :rtype: pt.Transformer
    :raises KeyError: if the text_attr or title_attr columns are not found in the input dataframe.
    
    Example::
    
        pipe = ( pt.terrier.Retriever(index, wmodel="DPH", metadata=["docno", "body"]) 
            >> pt.text.sliding(length=128, stride=64, prepend_attr=None) 
            >> pt.text.scorer(wmodel="DPH") 
            >> pt.text.max_passage() )

        # tokenizer model 
        from transformers import AutoTokenizer
        tok = AutoTokenizer.from_pretrained("bert-base-uncased")
        pipe = (pt.terrier.Retriever(index, wmodel="DPH", metadata=["docno", "body"])
            >> pt.text.sliding(length=128, stride=64, prepend_attr=None, tokenizer=tok)
            >> pt.text.scorer(wmodel="DPH")
            >> pt.text.max_passage() )
    """

    # deal with older names for attributes
    if 'passage_length' in kwargs:
        length = kwargs['passage_length']
        del kwargs['passage_length']
        warn(
            "passage_length should be length.", FutureWarning, 2)
    if 'passage_stride' in kwargs:
        stride = kwargs['passage_stride']
        del kwargs['passage_stride']
        warn(
            "passage_stride should be stride.", FutureWarning, 2)
    if 'prepend_title' in kwargs:
        warn(
            "prepend_title and title_attr should be replaced with prepend_attr.", FutureWarning, 2)
        if kwargs['prepend_title']:
            prepend_attr = kwargs['title_attr']
            del kwargs['title_attr']
        else:
            prepend_attr = None
        del kwargs['prepend_title']
        
    return SlidingWindowPassager(
        text_attr=text_attr, 
        passage_length=length, 
        passage_stride=stride,
        prepend_title=prepend_attr is not None,
        title_attr=prepend_attr,
        join=' ',
        tokenizer=tokenizer,
        **kwargs
    )

def max_passage() -> pt.Transformer:
    """
    Scores each document based on the maximum score of any constituent passage. Applied after a sliding window transformation
    has been scored.
    """
    return MaxPassage()

def mean_passage() -> pt.Transformer:
    """
    Scores each document based on the mean score of all constituent passages. Applied after a sliding window transformation
    has been scored.
    """
    return MeanPassage()

def first_passage() -> pt.Transformer:
    """
    Scores each document based on score of the first passage of that document. Note that this transformer is rarely used in conjunction with
    the sliding window transformer, as all passages would required to be scored, only for the first one to be used.
    """
    return FirstPassage()

def kmaxavg_passage(k : int) -> pt.Transformer:
    """
    Scores each document based on the average score of the top scoring k passages. Generalises combination of mean_passage()
    and max_passage(). Proposed in [Chen2020].

    :param k: The number of top-scored passages for each document to use when scoring
    
    """
    return KMaxAvgPassage(k)

def slidingWindow(sequence : list, winSize : int, step : int) -> list:
    """
    For the specified sequence, break into sliding windows of size winSize, 
    stepping forward by the specified amount each time

    :param sequence: the sequence to break into sliding windows
    :param winSize: the size of the sliding window
    :param step: how much to step forward each time
    :return: a list of sliding windows, where each window is a list of the elements in that window
    :rtype: list
    """
    return [x for x in list(more_itertools.windowed(sequence,n=winSize, step=step)) if x[-1] is not None]

def snippets(
        text_scorer_pipe : pt.Transformer, 
        text_attr : str = "text", 
        summary_attr : str = "summary", 
        num_psgs : int = 5, 
        joinstr : str ='...') -> pt.Transformer:
    """
    Applies query-biased summarisation (snippet), by applying the specified text scoring pipeline.
    Takes a return a dataframe with the columns ['qid', 'query', 'docno', text_attr], and returns a dataframe
    with the columns ['qid', 'query', 'docno', text_attr, summary_attr]. The summary_attr column contains the
    query-biased summary for that document, upto num_psgs passages, joined together with the specified joinstr.

    :param text_scorer_pipe: the pipeline for scoring passages in response to the query. Normally this applies passaging.
        The pipeline should take a dataframe with the columns ['qid', 'query', 'docno', text_attr] and return a dataframe
        with the columns ['qid', 'query', 'docno', text_attr, 'score', 'rank'], where these are smaller passages than the input df.
    :param text_attr: what is the name of the attribute that contains the text of the document
    :param summary_attr: what is the name of the attribute that should contain the query-biased summary for that document
    :param num_psgs: how many passages to select for the summary of each document
    :param joinstr: how to join passages for a given document together

    Example::

        # retrieve documents with text
        br = pt.terrier.Retriever(index, metadata=['docno', 'text'])

        # use Tf as a passage scorer on sliding window passages 
        psg_scorer = ( 
            pt.text.sliding(text_attr='text', length=15, prepend_attr=None) 
            >> pt.text.scorer(body_attr="text", wmodel='Tf', takes='docs')
        )
        
        # use psg_scorer for performing query-biased summarisation on docs retrieved by br 
        retr_pipe = br >> pt.text.snippets(psg_scorer)

    """
    tsp = (
        pt.apply.rename({'qid' : 'oldqid'}) 
        >> pt.apply.qid(lambda row: row['oldqid'] + '-' + row['docno']) 
        >> ( text_scorer_pipe % num_psgs )
        >> pt.apply.qid(drop=True)
        >> pt.apply.rename({'oldqid' : 'qid'})
    )

    def _qbsjoin(docres):
        if len(docres) == 0:
            docres[summary_attr] = pd.Series(dtype='str')
            return docres     

        psgres = tsp(docres)
        if len(psgres) == 0:
            print('no passages found in %d documents for query %s' % (len(docres), docres.iloc[0].query))
            docres = docres.copy()
            docres[summary_attr]  = ""
            return docres

        # dont assign new columns to an existing df
        psgres = psgres.copy()
        psgres[["olddocno", "pid"]] = psgres.docno.str.split("%p", expand=True)

        newdf = psgres.groupby(['qid', 'olddocno'])[text_attr].agg(joinstr.join).reset_index().rename(columns={text_attr : summary_attr, 'olddocno' : 'docno'})
        
        return docres.merge(newdf, on=['qid', 'docno'], how='left')
    rtr = pt.apply.generic(_qbsjoin, required_columns=['qid', 'query', 'docno', text_attr])
    rtr.subtransformers = types.MethodType(lambda self: {'tsp' : tsp}, rtr) # type: ignore[attr-defined]
    return rtr


class DePassager(pt.Transformer):

    def __init__(self, agg="max", **kwargs):
        super().__init__(**kwargs)
        self.agg = agg

    def transform(self, topics_and_res):
        pt.validate.columns(topics_and_res, includes=['qid', 'docno'] + (['score'] if self.agg != 'first' else []))
        topics_and_res = topics_and_res.copy()
        topics_and_res[["olddocno", "pid"]] = topics_and_res.docno.str.split("%p", expand=True) if len(topics_and_res) > 0 else pd.DataFrame(columns=["olddocno", "pid"])
        if self.agg == 'max':
            groups = topics_and_res.groupby(['qid', 'olddocno'])
            group_max_idx = groups['score'].idxmax()
            rtr = topics_and_res.loc[group_max_idx, :]
            rtr = rtr.drop(columns=['docno', 'pid']).rename(columns={"olddocno" : "docno"})
        
        if self.agg == 'first':
            #could this be done by just selectin pid = 0?
            topics_and_res.pid = topics_and_res.pid.astype(int)
            rtr = topics_and_res[topics_and_res.pid == 0].rename(columns={"olddocno" : "docno"})
            
            groups = topics_and_res.groupby(['qid', 'olddocno'])
            group_first_idx = groups['pid'].idxmin()
            rtr = topics_and_res.loc[group_first_idx, ]
            rtr = rtr.drop(columns=['docno', 'pid']).rename(columns={"olddocno" : "docno"})

        if self.agg == 'mean':
            rtr = topics_and_res.groupby(['qid', 'olddocno'])['score'].mean().reset_index().rename(columns={'olddocno' : 'docno'})
            #add query columns back
            rtr = rtr.merge(topics_and_res[pt.model.query_columns(topics_and_res)].drop_duplicates(), on='qid')

        if self.agg == 'kmaxavg':
            rtr = topics_and_res.groupby(['qid', 'olddocno'])['score'].apply(lambda ser: ser.nlargest(self.K).mean()).reset_index().rename(columns={'olddocno' : 'docno'})
            #add query columns back
            rtr = rtr.merge(topics_and_res[pt.model.query_columns(topics_and_res)].drop_duplicates(), on='qid')

        if "docid" in rtr.columns:
            rtr = rtr.drop(columns=['docid'])
        rtr = pt.model.add_ranks(rtr)
        return rtr

class KMaxAvgPassage(DePassager):
    """
        .. cite.dblp:: conf/trec/ChenHSCH020

        Usage:
            X >> SlidingWindowPassager() >>  Y >>  KMaxAvgPassage(2)
        where X is some kind of model for obtaining the text of documents and Y is a text scorer, such as BERT or ColBERT
    """
    def __init__(self, K, **kwargs):
        kwargs["agg"] = "kmaxavg"
        self.K = K
        super().__init__(**kwargs)

class MaxPassage(DePassager):
    def __init__(self, **kwargs):
        kwargs["agg"] = "max"
        super().__init__(**kwargs)

class FirstPassage(DePassager):
    def __init__(self, **kwargs):
        kwargs["agg"] = "first"
        super().__init__(**kwargs)

class MeanPassage(DePassager):
    def __init__(self, **kwargs):
        kwargs["agg"] = "mean"
        super().__init__(**kwargs)


class SlidingWindowPassager(pt.Transformer):
    schematic = {'label': 'SlidingWindow'}

    def __init__(self, text_attr='body', title_attr='title', passage_length=150, passage_stride=75, join=' ', prepend_title=True, tokenizer=None, **kwargs):
        super().__init__(**kwargs)
        self.text_attr=text_attr
        self.title_attr=title_attr
        self.passage_length = passage_length
        self.passage_stride= passage_stride
        self.join = join
        self.prepend_title = prepend_title
        self.tokenizer = tokenizer

        # check if the tokenizer has the `.tokenize()` and the `.convert_tokens_to_string()` method
        if self.tokenizer is not None:
            self.tokenize = self.tokenizer.tokenize
            self.detokenize = self.tokenizer.convert_tokens_to_string
        else:
            self.tokenize = re.compile(r"\s+").split
            self.detokenize = ' '.join

    def transform(self, topics_and_res):
        with pt.validate.any(topics_and_res) as v:
            cols = [self.text_attr] + ([self.title_attr] if self.prepend_title else [])
            v.result_frame(cols)
            v.document_frame(cols)
        print("calling sliding on df of %d rows" % len(topics_and_res))

        # now apply the passaging
        if "qid" in topics_and_res.columns: 
            return self.applyPassaging(topics_and_res, labels="label" in topics_and_res.columns)
        return self.applyPassaging_no_qid(topics_and_res)

    def applyPassaging_no_qid(self, df):
        result_columns = list(df.columns)
        if self.prepend_title:
            result_columns = [c for c in result_columns if c != self.title_attr]
        rows=[]
        for row in df.itertuples():
            row = row._asdict()
            toks = self.tokenize(row[self.text_attr])
            if len(toks) < self.passage_length:
                row['docno'] = row['docno'] + "%p0"
                row[self.text_attr] = self.detokenize(toks)
                if self.prepend_title:
                    row[self.text_attr] = str(row[self.title_attr]) + self.join + row[self.text_attr]
                    del(row[self.title_attr])
                rows.append(row)
            else:
                passageCount=0
                for i, passage in enumerate( slidingWindow(toks, self.passage_length, self.passage_stride)):
                    newRow = row.copy()
                    newRow['docno'] = row['docno'] + "%p" + str(i)
                    newRow[self.text_attr] = self.detokenize(passage)
                    if self.prepend_title:
                        newRow[self.text_attr] = str(row[self.title_attr]) + self.join + newRow[self.text_attr]
                        del(newRow[self.title_attr])
                    rows.append(newRow)
                    passageCount+=1
        return pd.DataFrame(rows, columns=result_columns)


    def applyPassaging(self, df, labels=True):
        newRows=[]
        labelCount=defaultdict(int)
        currentQid=None
        rank=0
        copy_columns=[]
        for col in ["score", "rank"]:
            if col in df.columns:
                copy_columns.append(col)

        if len(df) == 0:
            return pd.DataFrame(columns=['qid', 'query', 'docno', self.text_attr, 'score', 'rank'])
    
        with pt.tqdm('passsaging', total=len(df), desc='passaging', leave=False) as pbar:
            for row in df.itertuples(index=False):
                row = row._asdict()
                pbar.update(1)
                qid = row['qid']
                if currentQid is None or currentQid != qid:
                    rank=0
                    currentQid = qid
                rank+=1
                toks = self.tokenize(row[self.text_attr])
                if len(toks) < self.passage_length:
                    newRow = row.copy()
                    newRow['docno'] = row['docno'] + "%p0"
                    newRow[self.text_attr] = self.detokenize(toks)
                    if self.prepend_title:
                        newRow.drop(labels=[self.title_attr], inplace=True)
                        newRow[self.text_attr] = str(row[self.title_attr]) + self.join + newRow[self.text_attr]
                    if labels:
                        labelCount[row['label']] += 1
                    for col in copy_columns:
                        newRow[col] = row[col]
                    newRows.append(newRow)
                else:
                    passageCount=0
                    for i, passage in enumerate( slidingWindow(toks, self.passage_length, self.passage_stride)):
                        newRow = row.copy()
                        newRow['docno'] = row['docno'] + "%p" + str(i)
                        newRow[self.text_attr] = self.detokenize(passage)
                        if self.prepend_title:
                            #newRow.drop(labels=[self.title_attr], inplace=True)
                            del newRow[self.title_attr]
                            newRow[self.text_attr] = str(row[self.title_attr]) + self.join + newRow[self.text_attr]
                        for col in copy_columns:
                            newRow[col] = row[col]
                        if labels:
                            labelCount[row['label']] += 1
                        newRows.append(newRow)
                        passageCount+=1
        newDF = pd.DataFrame(newRows)
        newDF['query'] = newDF['query'].fillna('')
        newDF[self.text_attr] = newDF[self.text_attr].fillna('')
        newDF['qid'] = newDF['qid'].fillna('')
        newDF.reset_index(inplace=True,drop=True)
        return newDF



import os

import pandas as pd

import pyterrier as pt
from pandas.testing import assert_frame_equal
from .base import BaseTestCase


class TestTopicsParsing(BaseTestCase):
    def testSingleLine(self):
        topics = pt.io.read_topics(
            os.path.dirname(os.path.realpath(__file__)) + "/fixtures/singleline.topics",
            format="singleline",
        )
        self.assertEqual(2, len(topics))
        self.assertTrue("qid" in topics.columns)
        self.assertTrue("query" in topics.columns)
        self.assertEqual(topics["qid"][0], "1")
        self.assertEqual(topics["qid"][1], "2")
        self.assertEqual(topics["query"][0], "one")
        self.assertEqual(topics["query"][1], "two words")

    def test_parse_trec_topics_file_T(self):
        input = os.path.dirname(os.path.realpath(__file__)) + "/fixtures/topics.trec"
        exp_result = pd.DataFrame(
            [["1", "light"], ["2", "radiowave"], ["3", "sound"]],
            columns=["qid", "query"],
        )
        result = pt.io.read_topics(input)
        assert_frame_equal(exp_result, result)

    def test_parse_trec_topics_file_TREC1(self):
        input = os.path.dirname(os.path.realpath(__file__)) + "/fixtures/topics.51-100.gz"
        results = pt.io.read_topics(input)
        print(results)
        self.assertEqual(results.query("qid == '51'").iloc[0]["query"], "Airbus Subsidies")

    def test_parse_trec_topics_file_D(self):
        input = os.path.dirname(os.path.realpath(__file__)) + "/fixtures/topics.trec"
        exp_result = pd.DataFrame(
            [["1", "lights"], ["2", "radiowaves"], ["3", "sounds"]],
            columns=["qid", "query"],
        )
        result = pt.io.read_topics(
            input, format="trec", whitelist=["desc"], blacklist=["title"]
        )
        assert_frame_equal(exp_result, result)

    def test_parse_trecxml_topics_file(self):
        input = os.path.dirname(os.path.realpath(__file__)) + "/fixtures/topics.trecxml"
        result = pt.io.read_topics(input, format="trecxml", tags=["title"])
        exp_result = pd.DataFrame(
            [["1", "lights"], ["2", "radiowaves"], ["3", "sounds"]],
            columns=["qid", "query"],
        )
        assert_frame_equal(exp_result, result)



"""PyTerrier artifact module."""
import contextlib
import io
import json
import os
import sys
import tarfile
import tempfile
import requests
from datetime import datetime
from hashlib import sha256
from pathlib import Path
import typing
from typing import Any, Dict, Iterator, Optional, Tuple, Union, NamedTuple
from urllib.parse import ParseResult, urlparse

from lz4.frame import LZ4FrameFile

import pyterrier as pt

# NO_PATH is a sentinel value that is used to indicate that an artifact has no file path (e.g., it's in-memory).
class _NoPath:
    def __repr__(self):
        return 'Artifact.NO_PATH'
_NO_PATH = _NoPath()

class Artifact:
    """Base class for PyTerrier artifacts.

    An artifact is a component stored on disk, such as an index.

    Artifacts usually act as factories for transformers that use them. For example, an index artifact
    may provide a `.retriever()` method that returns a transformer that searches the index.
    """

    # expose NO_PATH sentinel
    NO_PATH = _NO_PATH

    def __init__(self, path: Union[Path, str, _NoPath]):
        """Initialize the artifact at the provided URL."""
        self.path: Union[Path, _NoPath] = path if isinstance(path, _NoPath) else Path(path)

    @classmethod
    def load(cls, path: str, **kwargs : Any) -> 'Artifact':
        """Load the artifact from the specified path.

        If invoked on the base class, this method will attempt to find a supporting Artifact
        implementation that can load the artifact at the specified path. If invoked on a subclass,
        it will attempt to load the artifact using the specific implementation.

        Args:
            path: The path of the artifact on disk.
            **kwargs: arguments that will be passed to the constructor of the artifact class 

        Returns:
            The loaded artifact.

        Raises:
            FileNotFoundError: If the specified path does not exist.
            ValueError: If no implementation is found that supports the artifact at the specified path.
        """
        if cls is Artifact:
            return _load(path)
        else:
            # called SomeArtifact.load(path), so load this specific artifact
            # TODO: add error message if not loaded
            return cls(path, **kwargs)

    @classmethod
    def from_url(cls, url: str, *, expected_sha256: Optional[str] = None, **kwargs : Any) -> 'Artifact':
        """Load the artifact from the specified URL.

        The artifact at the specified URL will be downloaded and stored in PyTerrier's artifact cache.

        Args:
            url: The URL or file path of the artifact.
            expected_sha256: The expected SHA-256 hash of the artifact. If provided, the downloaded artifact will be
                verified against this hash and an error will be raised if the hash does not match.
            **kwargs: arguments that will be passed to the constructor of the artifact class

        Returns:
            The loaded artifact.

        Raises:
            ValueError: If no implementation is found that supports the artifact at the specified path.
        """
        # Support mapping "protocols" of the URL other than http[s]
        # e.g., "hf:abc/xyz@branch" -> "https://huggingface.co/datasets/abc/xyz/resolve/branch/artifact.tar"
        parsed_url = urlparse(url)
        headers = {}
        protocol_entry_points = {ep.name: ep for ep in pt.utils.entry_points('pyterrier.artifact.url_protocol_resolver')}
        while parsed_url.scheme in protocol_entry_points:
            resolver = protocol_entry_points[parsed_url.scheme].load()
            resolver_result = resolver(parsed_url)
            del protocol_entry_points[parsed_url.scheme] # avoid the possiblity of an infinite loop here
            if resolver_result is not None:
                if isinstance(resolver_result, str):
                    resolver_result = UrlResolverResult(resolver_result) # url
                elif isinstance(resolver_result, tuple):
                    resolver_result = UrlResolverResult(*resolver_result)
                url = resolver_result.url
                parsed_url = urlparse(url)
                headers.update(resolver_result.headers)

        if parsed_url.scheme == '' and os.path.exists(url):
            return cls.load(url) # already resolved, load it

        # buid local path
        base_path = os.path.join(pt.io.pyterrier_home(), 'artifacts')
        if not os.path.exists(base_path):
            os.makedirs(base_path)
        path = os.path.join(base_path, sha256(url.encode()).hexdigest())

        if not os.path.exists(path):
            # try to load the package metadata file (if it exists)
            download_info = {}
            try:
                with pt.io.open_or_download_stream(f'{url}.json', headers=headers, verbose=False) as fin:
                    download_info = json.load(fin)
                if 'expected_sha256' in download_info:
                    new_expected_sha256 = download_info['expected_sha256'].lower()
                    if expected_sha256 is not None:
                        assert expected_sha256.lower() == new_expected_sha256
                    expected_sha256 = new_expected_sha256
                if 'total_size' in download_info:
                    # TODO: make sure there's available space on the disk for this
                    pass
            except OSError:
                pass


            with contextlib.ExitStack() as stack:
                dout = stack.enter_context(pt.io.finalized_directory(path))

                metadata_out = stack.enter_context(pt.io.finalized_open(f'{path}.json', 't'))

                if parsed_url.scheme == '':
                    info = {'path': url}
                else:
                    info = {'url': url}

                metadata_out.write(json.dumps(info) + '\n') # type: ignore[arg-type]

                if 'segments' in download_info:
                    # the file is segmented -- use a sequence reader to stitch them back together
                    fin = stack.enter_context(pt.io.MultiReader(
                        stack.enter_context(pt.io.open_or_download_stream(f'{url}.{i}', headers=headers))
                        for i in range(len(download_info['segments']))
                    ))
                    if expected_sha256 is not None:
                        fin = stack.enter_context(pt.io.HashReader(fin, expected=expected_sha256))
                else:
                    fin = stack.enter_context(pt.io.open_or_download_stream(url, headers=headers, expected_sha256=expected_sha256))

                cin = fin
                if parsed_url.path.endswith('.lz4'):
                    cin = stack.enter_context(LZ4FrameFile(fin))
                # TODO: support other compressions
                cin = typing.cast(io.BytesIO, cin) # noqa: PT100 (this is typing.cast, not jinus.cast)

                tar_in = stack.enter_context(tarfile.open(fileobj=cin, mode='r|'))

                metadata_out.flush()
                for member in tar_in:
                    if (member.isfile() or member.isdir()) and pt.io.path_is_under_base(member.path, dout):
                        print(f'extracting {member.path} [{pt.utils.byte_count_to_human_readable(member.size)}]')
                        tar_in.extract(member, dout, set_attrs=False)

        return cls.load(path)

    def _package_files(self) -> Iterator[Tuple[str, Union[str, io.BytesIO]]]:
        assert not isinstance(self.path, _NoPath), "package cannot be built for artifacts without a path"
        has_pt_meta_file = False
        for root, dirs, files in os.walk(self.path):
            rel_root = os.path.relpath(root, start=self.path)
            for file in sorted(files):
                file_full_path = os.path.join(root, file)
                file_rel_path = os.path.join(rel_root, file) if rel_root != '.' else file
                yield file_rel_path, file_full_path
                if file_rel_path == 'pt_meta.json':
                    has_pt_meta_file = True
        if not has_pt_meta_file:
            metadata = self._build_metadata()
            if metadata is not None:
                yield 'pt_meta.json', io.BytesIO(json.dumps(metadata).encode())

    def _build_metadata(self) -> Optional[Dict[str, Any]]:
        metadata = {}

        try:
            metadata['type'], metadata['format'] = pt.inspect.artifact_type_format(self) # type: ignore[misc]
        except TypeError:
            pass # couldn't identify type and format

        if hasattr(self, 'ARTIFACT_PACKAGE_HINT'):
            metadata['package_hint'] = self.ARTIFACT_PACKAGE_HINT
        elif self.__class__ is not Artifact:
            metadata['package_hint'] = self.__class__.__module__.split('.')[0]

        return metadata

    def build_package(
        self,
        package_path: Optional[str] = None,
        *,
        max_file_size: Optional[float] = None,
        metadata_out: Optional[Dict[str, Any]] = None,
        verbose: bool = True,
    ) -> str:
        """Builds a package for this artifact.

        Packaged artifacts are useful for distributing an artifact as a single file, such as from Artifact.from_url().
        A separate metadata file is also generated, which gives information about the package's contents, including
        file sizes and an expected hash for the package.

        Args:
            package_path: The path of the package to create. Defaults to the artifact path with a .tar.lz4 extension.
            max_file_size: the (approximate) maximum size of each file.
            metadata_out: A dictionary that is updated with the metadata of the artifact (if provided).
            verbose: Whether to display a progress bar when packaging.

        Returns:
            The path of the package created.
        """
        assert not isinstance(self.path, _NoPath), "package cannot be built for artifacts without a path"
        if package_path is None:
            package_path = str(self.path) + '.tar.lz4'

        metadata: Dict[str, Any] = {
            'expected_sha256': None,
            'total_size': 0,
            'contents': [],
        }

        chunk_num = 0
        chunk_start_offset = 0
        raw_fout: Optional[io.BufferedIOBase] = None
        def manage_maxsize(_: None):
            nonlocal raw_fout
            nonlocal chunk_num
            nonlocal chunk_start_offset
            if raw_fout is not None and max_file_size is not None and raw_fout.tell() >= max_file_size:
                raw_fout.flush()
                chunk_start_offset += raw_fout.tell()
                raw_fout.close()
                if chunk_num == 0:
                    metadata['segments'] = []
                    metadata['segments'].append({'idx': chunk_num, 'offset': 0})
                chunk_num += 1
                if verbose:
                    print(f'starting segment {chunk_num}')
                metadata['segments'].append({'idx': chunk_num, 'offset': chunk_start_offset})
                raw_fout = stack.enter_context(pt.io.finalized_open(f'{package_path}.{chunk_num}', 'b'))
                sha256_fout.replace_writer(raw_fout)

        with contextlib.ExitStack() as stack:
            raw_fout = stack.enter_context(pt.io.finalized_open(f'{package_path}.{chunk_num}', 'b'))
            sha256_fout = stack.enter_context(pt.io.HashWriter(raw_fout))
            lz4_fout = stack.enter_context(LZ4FrameFile(sha256_fout, 'wb'))
            tarout = stack.enter_context(tarfile.open(fileobj=lz4_fout, mode='w'))

            added_dirs = set()
            for rel_path, file in self._package_files():
                path_dir, name = os.path.split(rel_path)
                if path_dir and path_dir not in added_dirs:
                    tar_record = tarfile.TarInfo(path_dir)
                    tar_record.type = tarfile.DIRTYPE
                    tarout.addfile(tar_record)
                    added_dirs.add(path_dir)
                lz4_fout.flush() # flush before each file, allowing seeking directly to this file within the archive
                tar_record = tarfile.TarInfo(rel_path)
                if isinstance(file, io.BytesIO):
                    tar_record.size = file.getbuffer().nbytes
                else:
                    tar_record.size = os.path.getsize(file)
                metadata['contents'].append({
                    'path': rel_path,
                    'size': tar_record.size,
                    'offset': chunk_start_offset + raw_fout.tell(),
                })
                metadata['total_size'] += tar_record.size
                if verbose:
                    print(f'adding {rel_path} [{pt.utils.byte_count_to_human_readable(tar_record.size)}]')

                if isinstance(file, io.BytesIO):
                    file.seek(0)
                    if rel_path == 'pt_meta.json' and metadata_out is not None:
                        metadata_out.update(json.load(file))
                    file.seek(0)
                    with pt.io.CallbackReader(file, manage_maxsize) as fin:
                        tarout.addfile(tar_record, fin) # type: ignore[arg-type]
                else:
                    with open(file, 'rb') as _fin, \
                         pt.io.CallbackReader(_fin, manage_maxsize) as fin:
                        tarout.addfile(tar_record, fin) # type: ignore[arg-type]
                    if rel_path == 'pt_meta.json' and metadata_out is not None:
                        with open(file, 'rb') as _fin:
                            metadata_out.update(json.load(_fin))

            tarout.close()
            lz4_fout.close()

            metadata['expected_sha256'] = sha256_fout.hexdigest()

            metadata_outf = stack.enter_context(pt.io.finalized_open(f'{package_path}.json', 't'))
            metadata_outf.write(json.dumps(metadata) + '\n') # type: ignore[arg-type]

        if chunk_num == 0:
            # no chunking was actually done, can use provided name directly
            os.rename(f'{package_path}.{chunk_num}', package_path)

        return package_path

    @classmethod
    def from_dataset(cls, dataset: str, variant: str, *, expected_sha256: Optional[str] = None) -> 'Artifact':
        """Load an artifact from a PyTerrier dataset.

        Args:
            dataset: The name of the dataset.
            variant: The variant of the dataset.
            expected_sha256: The expected SHA-256 hash of the artifact. If provided, the downloaded artifact will be
                verified against this hash and an error will be raised if the hash does not match.
        """
        return cls.from_hf(
            repo='pyterrier/from-dataset',
            branch=f'{dataset}.{variant}',
            expected_sha256=expected_sha256)

    # -------------------------------------------------
    # HuggingFace Datasets Integration
    #  - from_hf()
    #  - to_hf()
    # -------------------------------------------------

    @classmethod
    def from_hf(cls, repo: str, branch: Optional[str] = None, *, expected_sha256: Optional[str] = None, **kwargs : Any) -> 'Artifact':
        """Load an artifact from Hugging Face Hub.

        Args:
            repo: The Hugging Face repository name.
            branch: The branch or tag of the repository to load. (Default: main). A branch can also be provided directly
                in the repository name using ``owner/repo@branch``.
            expected_sha256: The expected SHA-256 hash of the artifact. If provided, the downloaded artifact will be
                verified against this hash and an error will be raised if the hash does not match.
            **kwargs: arguments that will be passed to the constructor of the artifact class
        """
        if branch is not None:
            if '@' in repo:
                raise ValueError('Provided branch in both repository name (via @) and as argument to from_hf')
            repo = f'{repo}@{branch}'
        return cls.from_url(f'hf:{repo}', expected_sha256=expected_sha256, **kwargs)

    def to_hf(self,
        repo: str,
        *,
        branch: Optional[str] = None,
        pretty_name: Optional[str] = None,
        private: Optional[bool] = None,
    ) -> None:
        """Upload this artifact to Hugging Face Hub.

        Args:
            repo: The Hugging Face repository name.
            branch: The branch or tag of the repository to upload to. (Default: main) A branch can also be provided
                directly in the repository name using ``owner/repo@branch``.
            pretty_name: The human-readable name of the artifact. (Default: the repository name)
            private: Whether make the repository private. New repositories default to public unless the organization’s
                default is private. No change to the repository's visiblity will be made if ``private=None`` (default).
        """
        import huggingface_hub

        if '@' in repo:
            if branch is not None:
                raise ValueError('Provided branch in both repository name (via @) and as argument to to_hf')
            repo, branch = repo.split('@', 1)
        if branch is None:
            branch = 'main'

        with tempfile.TemporaryDirectory() as d:
            # build a package with a maximum individual file size of just under 5GB, the limit for HF datasets
            metadata: Dict[str, Any] = {}
            self.build_package(os.path.join(d, 'artifact.tar.lz4'), max_file_size=20e9, metadata_out=metadata)
            readme = self._hf_readme(repo=repo, branch=branch, pretty_name=pretty_name, metadata=metadata)
            if readme:
                with open(f'{d}/README.md', 'wt') as fout:
                    fout.write(readme)
            try:
                huggingface_hub.create_repo(repo, repo_type='dataset', private=private)
            except huggingface_hub.utils.HfHubHTTPError as e:
                if e.server_message != 'You already created this dataset repo':
                    raise

            if private is not None:
                huggingface_hub.update_repo_visibility(repo, repo_type='dataset', private=private)

            try:
                huggingface_hub.create_branch(repo, repo_type='dataset', branch=branch)
            except huggingface_hub.utils.HfHubHTTPError as e:
                # detect the "already exists" error -- either as the 409 (Conflict) status code or 'already exists' appearing in the message
                if (hasattr(e, 'response') and e.response.status_code == 409) or 'already exists' in str(e.server_message).lower():
                    pass # ignore -- the branch already exists
                else:
                    raise # another error

            path = huggingface_hub.upload_folder(
                repo_id=repo,
                folder_path=d,
                repo_type='dataset',
                revision=branch,
            )
            sys.stderr.write(f"\nArtifact uploaded to {path}\nConsider editing the README.md to help explain this "
                              "artifact to others.\n")

    def _hf_readme(self,
        *,
        repo: str,
        branch: Optional[str] = 'main',
        pretty_name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None
    ) -> Optional[str]:
        if pretty_name is None:
            title = repo.split('/')[-1]
            pretty_name = '# pretty_name: "" # Example: "MS MARCO Terrier Index"'
        else:
            title = pretty_name
            pretty_name = f'pretty_name: {pretty_name!r}'
        if branch != 'main':
            repo = f'{repo}@{branch}'
        if metadata is None:
            metadata = {}
        tags = ['pyterrier', 'pyterrier-artifact']
        if 'type' in metadata:
            tags.append('pyterrier-artifact.{type}'.format(**metadata))
        if 'type' in metadata and 'format' in metadata:
            tags.append('pyterrier-artifact.{type}.{format}'.format(**metadata))
        stags = '\n- '.join([''] + tags)
        return f'''---
{pretty_name}
tags:{stags}
task_categories:
- text-retrieval
viewer: false
---

# {title}

## Description

*TODO: What is the artifact?*

## Usage

```python
# Load the artifact
import pyterrier as pt
artifact = pt.Artifact.from_hf({repo!r})
# TODO: Show how you use the artifact
```

## Benchmarks

*TODO: Provide benchmarks for the artifact.*

## Reproduction

```python
# TODO: Show how you constructed the artifact.
```

## Metadata

```
{json.dumps(metadata, indent=2)}
```
'''

    # -------------------------------------------------
    # Zenodo Integration
    #  - from_zenodo()
    #  - to_zenodo()
    # -------------------------------------------------

    @classmethod
    def from_zenodo(cls, zenodo_id: str, *, expected_sha256: Optional[str] = None, **kwargs : Any) -> 'Artifact':
        """Load an artifact from Zenodo.

        Args:
            zenodo_id: The Zenodo record ID of the artifact.
            expected_sha256: The expected SHA-256 hash of the artifact. If provided, the downloaded artifact will be
                verified against this hash and an error will be raised if the hash does not match.
            **kwargs: arguments that will be passed to the constructor of the artifact class
        """
        return cls.from_url(f'zenodo:{zenodo_id}', expected_sha256=expected_sha256, **kwargs)

    def to_zenodo(self, *, pretty_name: Optional[str] = None, sandbox: bool = False) -> None:
        """Upload this artifact to Zenodo.

        Args:
            pretty_name: The human-readable name of the artifact.
            sandbox: Whether to perform a test upload to the Zenodo sandbox.
        """
        if sandbox:
            base_url = 'https://sandbox.zenodo.org/api'
        else:
            base_url = 'https://zenodo.org/api'

        access_token = os.environ.get('ZENODO_TOKEN')
        params = {'access_token': access_token}

        with tempfile.TemporaryDirectory() as d:
            r = requests.post(f'{base_url}/deposit/depositions', params=params, json={})
            r.raise_for_status()
            deposit_data = r.json()
            sys.stderr.write("Created {}\n".format(deposit_data['links']['html']))
            try:
                metadata: Dict[str, Any] = {}
                sys.stderr.write("Building package.\n")
                self.build_package(os.path.join(d, 'artifact.tar.lz4'), metadata_out=metadata)
                z_meta = {
                    'metadata': self._zenodo_metadata(
                        pretty_name=pretty_name,
                        zenodo_id=deposit_data['id'],
                        metadata=metadata,
                    ),
                }
                r = requests.put(deposit_data['links']['latest_draft'], params=params, json=z_meta)
                r.raise_for_status()
                sys.stderr.write("Uploading...\n")
                for file in sorted(os.listdir(d)):
                    file_path = os.path.join(d, file)
                    with open(file_path, 'rb') as _fin, \
                         pt.io.TqdmReader(_fin, total=os.path.getsize(file_path), desc=file) as fin:
                        r = requests.put(
                            '{}/{}'.format(deposit_data['links']['bucket'], file),
                            params={'access_token': access_token},
                            data=fin)
                        r.raise_for_status()
            except:
                sys.stderr.write("Discarding {}\n".format(deposit_data['links']['html']))
                requests.post(deposit_data['links']['discard'], params=params, json={})
                raise
            sys.stderr.write("Upload complete. Please complete the form at {} to publish this artifact. (Note that "
                "publishing to Zenodo cannot be undone.)\n".format(deposit_data['links']['html']))

    def _zenodo_metadata(self, *, zenodo_id: str, pretty_name: Optional[str] = None, metadata: Dict) -> Dict[str, Any]:
        description = f'''
<h2>Description</h2>

<p>
<i>TODO: What is the artifact?</i>
</p>

<h2>Usage</h2>

<pre>
# Load the artifact
import pyterrier as pt
artifact = pt.Artifact.from_zenodo({str(zenodo_id)!r})
# TODO: Show how you use the artifact
</pre>

<h2>Benchmarks</h2>

<p>
<i>TODO: Provide benchmarks for the artifact.</i>
</p>

<h2>Reproduction</h2>

<pre>
# TODO: Show how you constructed the artifact.
</pre>

<h2>Metadata</h2>

<pre>
{json.dumps(metadata, indent=2)}
</pre>
'''
        tags = ['pyterrier', 'pyterrier-artifact']
        if 'type' in metadata:
            tags.append('pyterrier-artifact.{type}'.format(**metadata))
        if 'type' in metadata and 'format' in metadata:
            tags.append('pyterrier-artifact.{type}.{format}'.format(**metadata))
        metadata = {
            'description': description,
            'upload_type': 'other',
            'publisher': 'Zenodo',
            'publication_date': datetime.today().strftime('%Y-%m-%d'),
            'keywords': tags,
        }
        if pretty_name:
            metadata['title'] = pretty_name
        return metadata

    # -------------------------------------------------
    # Magic Wormhole Integration
    #  - from_p2p()
    #  - to_p2p()
    # -------------------------------------------------

    @classmethod
    def from_p2p(cls, code: str, path: str, *, expected_sha256: Optional[str] = None, **kwargs : Any) -> 'Artifact':
        """Load an artifact from a peer using a Magic Wormhole code.

        Args:
            code: The Magic Wormhole code.
            path: The path to save the artifact to.
            expected_sha256: The expected SHA-256 hash of the artifact. If provided, the downloaded artifact will be
                verified against this hash and an error will be raised if the hash does not match.
            **kwargs: arguments that will be passed to the constructor of the artifact class
        """
        import wormhole # noqa check that magic-wormhole is installed
        import subprocess
        if os.path.exists(path):
            raise FileExistsError(f'{path!r} already exists. Choose a different path.')
        with tempfile.TemporaryDirectory() as d:
            tmp_path = os.path.join(d, 'artifact.tar.lz4')
            command = [sys.executable, '-m', 'wormhole', 'receive', code, '-o', tmp_path, '--accept-file']
            subprocess.run(command)
            with LZ4FrameFile(tmp_path) as fin, \
                 tarfile.open(fileobj=fin, mode='r|') as tar_in, \
                 pt.io.finalized_directory(path) as dout:
                for member in tar_in:
                    if (member.isfile() or member.isdir()) and pt.io.path_is_under_base(member.path, dout):
                        print(f'extracting {member.path} [{pt.utils.byte_count_to_human_readable(member.size)}]')
                        tar_in.extract(member, dout, set_attrs=False)
        return cls.load(path)

    def to_p2p(self):
        """Send this artifact directly to a peer using Magic Wormhole.

        The recipient can use the provided code to download the artifact.
        """
        import wormhole # noqa check that magic-wormhole is installed
        import subprocess
        with tempfile.TemporaryDirectory() as d:
            path = os.path.join(d, 'artifact.tar.lz4')
            self.build_package(path)
            command = [sys.executable, '-m', 'wormhole', 'send', path]
            try:
                with subprocess.Popen(command, stderr=subprocess.PIPE, text=True) as process:
                    for line in process.stderr:
                        if line.startswith('Wormhole code is'):
                            code = line.replace('Wormhole code is:', '').strip()
                            name = Path(self.path).name
                            print()
                            print('Ready to send. Run the following on the target machine (ctrl+c to cancel):')
                            print('import pyterrier as pt')
                            print(f'artifact = pt.Artifact.from_p2p({code!r}, {name!r})')
                        elif any(x in line for x in ['wormhole receive', 'On the other computer', 'Sending ']) or line.strip() == '':
                            pass
                        else:
                            print(line, end='')
            except KeyboardInterrupt:
                pass


def _load_metadata(path: str) -> Dict:
    """Load the metadata file for the artifact at the specified path.

    Args:
        path: The path of the artifact on disk.

    Returns:
        The loaded metadata file, or an empty dictionary if the metadata file does not exist.

    Raises:
        FileNotFoundError: If the specified path does not exist.
    """
    if not os.path.isdir(path):
        if os.path.isfile(path):
            return {}
        raise FileNotFoundError(f'{path} not found')

    metadata_file = os.path.join(path, 'pt_meta.json')
    if os.path.exists(metadata_file):
        with open(metadata_file, 'rt') as f:
            return json.load(f)

    return {}


def _metadata_adapter(path: str):
    directory_listing = []
    if os.path.isdir(path):
        directory_listing = os.listdir(path)
    for entry_point in pt.utils.entry_points('pyterrier.artifact.metadata_adapter'):
        if (metadata := entry_point.load()(path, directory_listing)) is not None:
            return metadata
    return {}


def _load(path: str, **kwargs : Any) -> Artifact:
    """Load the artifact from the specified path.

    This function attempts to load the artifact using a supporting Artifact implementation.

    Args:
        path: The path of the artifact on disk.
        **kwargs: arguments that will be passed to the constructor of the artifact class

    Returns:
        The loaded artifact.

    Raises:
        FileNotFoundError: If the specified path does not exist.
        ValueError: If no implementation is found that supports the artifact at the specified path.
    """
    metadata = _load_metadata(path)

    if 'type' not in metadata or 'format' not in metadata:
        metadata.update(_metadata_adapter(path))

    if 'type' in metadata and 'format' in metadata:
        entry_points = list(pt.utils.entry_points('pyterrier.artifact'))
        matching_entry_points = [ep for ep in entry_points if ep.name == '{type}.{format}'.format(**metadata)]
        for entry_point in matching_entry_points:
            artifact_cls = entry_point.load()
            return artifact_cls(path, **kwargs)

    # coudn't find an implementation that supports this artifact
    error = [
        f'No implementation found that supports the artifact at {_path_repr(path)}.',
        f'type={metadata.get("type", "(unknown)")}, format={metadata.get("format", "(unknown)")}.'
    ]
    if 'package_hint' in metadata:
        error.append(f'Do you need to `pip install {metadata["package_hint"]}`?')
    raise ValueError('\n'.join(error))


def _path_repr(path: str) -> str:
    """Return a string representation of a path, including information from the artifact's metadata file."""
    artifact_info_path = f'{path}.json'
    if os.path.exists(artifact_info_path):
        with open(artifact_info_path) as fin:
            artifact_info = json.load(fin)
        if 'url' in artifact_info:
            return f'{path!r} <from {artifact_info["url"]!r}>'
        if 'path' in artifact_info:
            return f'{path!r} <from {artifact_info["path"]!r}>'
    return repr(path)


class UrlResolverResult(NamedTuple):
    url: str
    headers: Dict[str, str] = {}


def _hf_url_resolver(parsed_url: ParseResult) -> Optional[UrlResolverResult]:
    # paths like: hf:macavaney/msmarco-passage.terrier
    org_repo = parsed_url.path
    # default to ref=main, but allow user to specify another branch, hash, etc, with abc/xyz@branch
    ref = 'main'
    if '@' in org_repo:
        org_repo, ref = org_repo.split('@', 1)
    headers = {}
    try:
        # try to get the token using huggingface_hub (takes from various sources: HF_TOKEN env, file, etc.)
        import huggingface_hub.utils
        token = huggingface_hub.utils.get_token()
    except ImportError:
        # since huggingface_hub isn't a required dependency, at least try to take from HF_TOKEN as a fallback
        token = os.environ.get('HF_TOKEN')
    if token:
        headers['Authorization'] = f'Bearer {token}'
    return UrlResolverResult(f'https://huggingface.co/datasets/{org_repo}/resolve/{ref}/artifact.tar.lz4', headers)


def _zenodo_url_resolver(parsed_url: ParseResult) -> str:
    # paths like: zenodo:111952
    zenodo_id = parsed_url.path
    return f'https://zenodo.org/records/{zenodo_id}/files/artifact.tar.lz4'



from pyterrier.java import mavenresolver
from pyterrier.java._utils import JavaInitializer, init, started, parallel_init, parallel_init_args, required, required_raise, before_init, autoclass, cast, legacy_init, register_config, Configuration, JavaClasses
from pyterrier.java._core import ColabJavaInit, CoreJavaInit, Java24Init, configure, redirect_stdouterr, bytebuffer_to_array, J, add_jar, add_package, set_memory_limit, set_redirect_io, add_option, set_log_level, set_java_home

__all__ = [
    'add_jar', 'add_option', 'add_package', 'autoclass', 'before_init', 'bytebuffer_to_array', 'cast', 'config',
    'configure', 'init', 'J', 'JavaClasses', 'mavenresolver', 'parallel_init', 'parallel_init_args',
    'redirect_stdouterr', 'required', 'required_raise', 'set_log_level', 'set_memory_limit', 'set_redirect_io',
    'started', 'legacy_init', 'JavaInitializer', 'ColabJavaInit', 'CoreJavaInit', 'Java24Init', 'set_java_home', 'register_config', 
    'Configuration',
]



import pandas as pd
import pyterrier as pt
from .base import BaseTestCase
import re


class TestText(BaseTestCase):

    def test_scorer(self):
        dfIn = pd.DataFrame(
            [
                ["q1", "chemical reactions", "d1", "professor protor poured the chemicals"],
                ["q1", "chemical reactions", "d2", "chemical brothers turned up the beats"],
            ], columns=["qid", "query", "docno", "text"])
        dfOut = pt.text.scorer(body_attr="text", wmodel="Tf").transform(dfIn)
        self.assertTrue("rank" in dfOut.columns)
        self.assertTrue("score" in dfOut.columns)
        self.assertEqual(1.0, dfOut.iloc[0].score)
        self.assertEqual(1.0, dfOut.iloc[1].score)

        # now check that text scorer can also work on an empty dataframe
        dfEmpty = pt.text.scorer(body_attr="text", wmodel="Tf").transform(dfIn.head(0))
        self.assertEqual(0, len(dfEmpty))
        self.assertIn("rank", dfEmpty.columns)
        self.assertIn("score", dfEmpty.columns)

    def test_scorer_rerank(self):
        #checks that the rank attribute is updated.
        dfIn = pd.DataFrame(
            [
                ["q1", "chemical reactions", "d1", 1.0, 0, "professor protor poured the chemicals"],
                ["q1", "chemical reactions", "d2", 0.9, 1, "chemical chemical chemical brothers turned up the beats"],
            ], columns=["qid", "query", "docno", "score", "rank", "text"])
        dfOut = pt.text.scorer(body_attr="text", wmodel="Tf").transform(dfIn)
        self.assertTrue("rank" in dfOut.columns)
        self.assertTrue("score" in dfOut.columns)
        self.assertEqual("d1", dfOut.iloc[0].docno)
        self.assertEqual(1.0, dfOut.iloc[0].score)
        self.assertEqual(3.0, dfOut.iloc[1].score)
        self.assertEqual(0, dfOut.iloc[1]["rank"])
        self.assertEqual(1, dfOut.iloc[0]["rank"])

    def test_snippets(self):
        br = pt.terrier.Retriever(self.here + "/fixtures/index/data.properties") >> pt.text.get_text(pt.get_dataset('irds:vaswani'), "text")
        #br = pt.terrier.Retriever.from_dataset("vaswani", "terrier_stemmed_text", metadata=["docno", "text"])
        psg_scorer = ( 
            pt.text.sliding(text_attr='text', length=25, stride=12, prepend_attr=None) 
            >> pt.text.scorer(body_attr="text", wmodel='Tf', takes='docs')
        )
        pipe = br >> pt.text.snippets(psg_scorer)
        dfOut = pipe.search("chemical reactions")
        self.assertTrue("rank" in dfOut.columns)
        self.assertTrue("score" in dfOut.columns)
        self.assertTrue("summary" in dfOut.columns)
        #.count() checks number of non-NaN values
        #so lets count how many are NaN
        self.assertEqual(0, len(dfOut) - dfOut.summary.count())


    def test_fetch_text_docno(self):
        dfinput = pd.DataFrame([["q1", "a query", "1"]], columns=["qid", "query", "docno"])
        #directory, indexref, str, Index
        for indexlike in [
            pt.get_dataset("vaswani").get_index(), 
            pt.IndexRef.of(pt.get_dataset("vaswani").get_index()),
            pt.IndexRef.of(pt.get_dataset("vaswani").get_index()).toString(),
            pt.IndexFactory.of(pt.get_dataset("vaswani").get_index())
        ]:
            textT = pt.text.get_text(indexlike, "docno")
            self.assertTrue(isinstance(textT, pt.transformer.Transformer))
            dfOut = textT.transform(dfinput)
            self.assertTrue(isinstance(dfOut, pd.DataFrame))
            self.assertTrue("docno" in dfOut.columns)

            # verify that rankcutoff is moved earlier by compilation
            pipe = textT%10
            pipe_opt = pipe.compile()
            self.assertEqual(textT, pipe_opt[1])
            self.assertIsInstance(pipe_opt[0], pt._ops.RankCutoff)
            self.assertEqual(10, pipe_opt[0].k)
        
    def test_fetch_text_docid(self):
        dfinput = pd.DataFrame([["q1", "a query", 1]], columns=["qid", "query", "docid"])
        df_empty = dfinput.head(0)
        #directory, indexref, str, Index
        for indexlike in [
            pt.get_dataset("vaswani").get_index(), 
            pt.IndexRef.of(pt.get_dataset("vaswani").get_index()),
            pt.IndexRef.of(pt.get_dataset("vaswani").get_index()).toString(),
            pt.IndexFactory.of(pt.get_dataset("vaswani").get_index())
        ]:
            textT = pt.text.get_text(indexlike, "docno")
            self.assertTrue(isinstance(textT, pt.Transformer))
            dfOut = textT.transform(dfinput)
            self.assertTrue(isinstance(dfOut, pd.DataFrame))
            self.assertTrue("docno" in dfOut.columns)
            self.assertEqual('object', dfOut['docno'].dtype)

            dfOut2 = textT.transform(df_empty)
            self.assertTrue(isinstance(dfOut2, pd.DataFrame))
            self.assertTrue("docno" in dfOut2.columns)
            self.assertEqual('object', dfOut2['docno'].dtype)

    def test_fetch_text_irds(self):
        dfinput = pd.DataFrame([
            ["q1", "a query", "4"],
            ["q1", "a query", "1"],
            ["q1", "a query", "4"],
            ], columns=["qid", "query", "docno"])
        df_empty = dfinput.head(0)
        textT = pt.text.get_text(pt.get_dataset('irds:vaswani'), "text")
        self.assertTrue(isinstance(textT, pt.Transformer))
        dfOut = textT.transform(dfinput)
        self.assertTrue(isinstance(dfOut, pd.DataFrame))
        self.assertTrue("text" in dfOut.columns)
        self.assertEqual('object', dfOut['docno'].dtype)
        self.assertTrue("the british computer society  report of a conference held in cambridge\njune\n" in dfOut.iloc[0].text)
        self.assertTrue("compact memories have flexible capacities  a digital data storage\nsystem with capacity up to bits and random and or sequential access\nis described\n" in dfOut.iloc[1].text)
        self.assertTrue("the british computer society  report of a conference held in cambridge\njune\n" in dfOut.iloc[2].text)

        dfOut2 = textT.transform(df_empty)
        self.assertTrue(isinstance(dfOut2, pd.DataFrame))
        self.assertTrue("text" in dfOut2.columns)
        self.assertEqual('object', dfOut2['docno'].dtype)

        # verify that rankcutoff is moved earlier by compilation
        pipe = textT%10
        pipe_opt = pipe.compile()
        self.assertEqual(textT, pipe_opt[1])
        self.assertIsInstance(pipe_opt[0], pt._ops.RankCutoff)
        self.assertEqual(10, pipe_opt[0].k)

    def test_passager_title(self):
        dfinput = pd.DataFrame([["q1", "a query", "doc1", "title", "body sentence"]], columns=["qid", "query", "docno", "title", "body"])
        passager = pt.text.sliding(length=1, stride=1)
        dfoutput = passager(dfinput)
        self.assertEqual(2, len(dfoutput))
        self.assertEqual("doc1%p0", dfoutput["docno"][0])
        self.assertEqual("doc1%p1", dfoutput["docno"][1])
        
        self.assertEqual("title body", dfoutput["body"][0])
        self.assertEqual("title sentence", dfoutput["body"][1])

    def test_passager(self):
        dfinput = pd.DataFrame([["q1", "a query", "doc1", "body sentence"]], columns=["qid", "query", "docno", "body"])
        passager = pt.text.sliding(length=1, stride=1, prepend_attr=None)
        dfoutput = passager(dfinput)
        self.assertEqual(2, len(dfoutput))
        self.assertEqual("doc1%p0", dfoutput["docno"][0])
        self.assertEqual("doc1%p1", dfoutput["docno"][1])
        
        self.assertEqual("body", dfoutput["body"][0])
        self.assertEqual("sentence", dfoutput["body"][1])

    def test_passager_custom_tokenizer(self):
        class MockTokenizer:
            def tokenize(self, input_str):
                rx = r"\w+(?:\"\w+)?|[^\w\s]"
                return re.findall(rx, input_str)
            
            def convert_tokens_to_string(self, input_toks):
                return ' '.join(input_toks)
    
        dfinput = pd.DataFrame([["q1", "a query", "doc1", "it's a sample document!"]], columns=["qid", "query", "docno", "body"])
        passager = pt.text.sliding(length=4, stride=3, prepend_attr=None, tokenizer=MockTokenizer())
        dfoutput = passager(dfinput)
        self.assertEqual(2, len(dfoutput))
        self.assertEqual("doc1%p0", dfoutput["docno"][0])
        self.assertEqual("doc1%p1", dfoutput["docno"][1])

        self.assertEqual("it ' s a", dfoutput["body"][0])
        self.assertEqual("a sample document !", dfoutput["body"][1])

    def test_passager_HGF(self):
        try: 
            from transformers import AutoTokenizer
        except:
            self.skipTest("`transformers` not installed")
        
        tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased")
        dfinput = pd.DataFrame([["q1", "a query", "doc1", "it's a sample document!"]], columns=["qid", "query", "docno", "body"])
        passager = pt.text.sliding(length=4, stride=3, prepend_attr=None, tokenizer=tokenizer)
        dfoutput = passager(dfinput)
        self.assertEqual(2, len(dfoutput))
        self.assertEqual("doc1%p0", dfoutput["docno"][0])
        self.assertEqual("doc1%p1", dfoutput["docno"][1])

        self.assertEqual("it ' s a", dfoutput["body"][0])
        self.assertEqual("a sample document!", dfoutput["body"][1])


    def test_depassager(self):
        dfinput = pd.DataFrame([["q1", "a query", "doc1", 1, "title", "body sentence"]], columns=["qid", "query", "docno", "docid", "title", "body"])
        passager = pt.text.sliding(length=1, stride=1)
        dfpassage = passager(dfinput)
        #     qid    query    docno            body                                         
        # 0  q1  a query  doc1%p0      title body
        # 1  q1  a query  doc1%p1  title sentence
        dfpassage["score"] = [1, 0]

        dfmax = pt.text.max_passage()(dfpassage)
        self.assertEqual(1, len(dfmax))
        self.assertEqual(1, dfmax["score"][0])
        self.assertTrue("query" in dfmax.columns)
        self.assertTrue("body" in dfmax.columns)
        self.assertFalse("docid" in dfmax.columns)
        self.assertFalse("pid" in dfmax.columns)
        self.assertFalse("olddocno" in dfmax.columns)
        self.assertEqual("a query", dfmax["query"][0])

        dffirst = pt.text.first_passage()(dfpassage)
        self.assertEqual(1, len(dffirst))
        self.assertEqual(1, dffirst["score"][0])
        self.assertTrue("body" in dffirst.columns)
        self.assertFalse("pid" in dffirst.columns)
        self.assertFalse("docid" in dffirst.columns)
        self.assertFalse("olddocno" in dffirst.columns)

        dfmean = pt.text.mean_passage()(dfpassage)
        self.assertEqual(1, len(dfmean))
        self.assertEqual(0.5, dfmean["score"][0])
        self.assertFalse("pid" in dfmean.columns)
        self.assertFalse("docid" in dfmean.columns)
        self.assertFalse("olddocno" in dfmax.columns)

        dfmeanK = pt.text.kmaxavg_passage(2)(dfpassage)
        self.assertEqual(1, len(dfmeanK))
        self.assertEqual(0.5, dfmeanK["score"][0])
        self.assertFalse("pid" in dfmeanK.columns)
        self.assertFalse("docid" in dfmeanK.columns)
        self.assertFalse("olddocno" in dfmeanK.columns)

        dfscores = pd.DataFrame([
            ["0", "doc1%p0", 3, pt.model.FIRST_RANK],
            ["0", "doc1%p1", 2, pt.model.FIRST_RANK+1],
            ["0", "doc1%p2", 1, pt.model.FIRST_RANK+2],
        ], columns=['qid', 'docno', 'score', 'rank'])

        dfmeanK2 = pt.text.kmaxavg_passage(2)(dfscores)
        print(dfmeanK2)
        self.assertEqual(1, len(dfmeanK2))
        self.assertEqual(2.5, dfmeanK2["score"][0])

        dfmeanK3 = pt.text.kmaxavg_passage(3)(dfscores)
        self.assertEqual(1, len(dfmeanK3))
        self.assertEqual(2, dfmeanK3["score"][0])

@pt.testing.transformer_test_class
def test_text_sliding():
    return pt.text.sliding()

@pt.testing.transformer_test_class
def test_max_passage():
    return pt.text.max_passage()



from docutils import nodes
from sphinx.directives.code import CodeBlock
from sphinx.application import Sphinx
from docutils.statemachine import StringList
import re


class CodeBlockFootnotes(CodeBlock):
    """
    Enhanced code-block directive that extracts footnotes from comments.
    """
    
    def run(self):
        # Get the code content
        code = '\n'.join(self.content)
        
        # Find all footnotes in the code
        footnotes = []
        footnote_pattern = r'#\s*:footnote:\s*(.+?)(?=\n|$)'
        
        def replace_footnote(match):
            footnote_text = match.group(1).strip()
            footnote_num = len(footnotes) + 1
            if ':nocomment:' in footnote_text:
                footnote_text = footnote_text.replace(':nocomment:', '').strip()
                comment = ''
            else:
                comment = f'# [{footnote_num}]'
            footnotes.append(footnote_text)
            return comment
        
        # Replace footnotes in code with numbered references
        modified_code = re.sub(footnote_pattern, replace_footnote, code)
        
        # Update the content with modified code
        self.content = modified_code.split('\n')
        
        # Run the parent CodeBlock directive to create the code block
        code_nodes = super().run()
        
        # If there are footnotes, add them after the code block
        if footnotes:
            # Create a container for the code block and footnotes
            container = nodes.container()
            container += code_nodes
            
            # Create footnote list
            footnote_list = nodes.enumerated_list()
            footnote_list['classes'].append('code-footnotes')
            
            for footnote_text in footnotes:
                list_item = nodes.list_item()
                
                # Parse the footnote text as reStructuredText
                self.state.nested_parse(
                    StringList([footnote_text], source=''),
                    self.content_offset,
                    list_item
                )
                
                footnote_list += list_item
            
            container += footnote_list
            
            return [container]
        
        return code_nodes


def setup(app: Sphinx):
    """
    Setup function for the Sphinx extension.
    """
    # Override the default code-block directive
    app.add_directive('code-block', CodeBlockFootnotes, override=True)
    
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import pandas as pd
import pyterrier as pt
import unittest
import os
from .base import TempDirTestCase
from glob import glob

class TestUtils(TempDirTestCase):

    def test_compression(self):
        with pt.io.autoopen(os.path.join(self.here, 'fixtures/testfile.txt.lz4'), 'rt') as f:
            lines = f.readlines()
            self.assertEqual('hello world', lines[0])
        with pt.io.autoopen(os.path.join(self.test_dir, "bla.txt.gz"), 'wt') as f:
            f.write('bla')
        with pt.io.autoopen(os.path.join(self.test_dir, "bla.txt.gz"), 'rt') as f:
            lines = f.readlines()
            self.assertEqual('bla', lines[0])

    def test_save_trec(self):
        res = pd.DataFrame([["1", "d1", 5.3, 1]], columns=['qid', 'docno', 'score', 'rank'])
        res_dict = res.set_index(['qid', 'docno']).to_dict()
        for filename in ["rtr.res", "rtr.res.gz", "rtr.res.bz2"]:
            filepath = os.path.join(self.test_dir, filename)
            pt.io.write_results(res, filepath, format="trec")
            res2 = pt.io.read_results(filepath)
            res2_dict = res2.set_index(['qid', 'docno']).to_dict()
            del res2_dict["name"]
            self.assertEqual(res_dict, res2_dict)

    def test_save_trec_generator(self):
        br = pt.terrier.Retriever(pt.get_dataset("vaswani").get_index(), wmodel="TF_IDF")
        filepath = os.path.join(self.test_dir, "test.res")
        pt.io.write_results(br.transform_gen(pt.get_dataset("vaswani").get_topics().head()), filepath, format="trec")

    def test_save_letor(self):
        import numpy as np
        res = pd.DataFrame([["1", "d1", 5.3, 1, np.array([1, 2])], ["1", "d2", 5.3, 1, np.array([2, 1])]], columns=['qid', 'docno', 'score', 'rank', 'features'])
        res_dict = res.set_index(['qid', 'docno']).to_dict()
        del res_dict["score"]
        del res_dict["rank"]
        for filename in ["rtr.letor", "rtr.letor.gz", "rtr.letor.bz2"]:
            filepath = os.path.join(self.test_dir, filename)
            pt.io.write_results(res, filepath, format="letor")
            res2 = pt.io.read_results(filepath, format="letor")

            for ((i1, row1), (i2, row2)) in zip(res.iterrows(), res2.iterrows()):
                self.assertEqual(row1["qid"], row2["qid"])
                self.assertEqual(row1["docno"], row2["docno"])
                self.assertEqual(row1["docno"], row2["docno"])
                self.assertEqual(row1["qid"], row2["qid"])
                self.assertTrue(np.array_equal(row1["features"], row2["features"]))

    def test_finalized_open(self):
        self.assertFalse(os.path.exists('file.txt'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.txt'))))

        try:
            with pt.io.finalized_open('file.txt', 't') as f:
                f.write('OK')
                self.assertFalse(os.path.exists('file.txt'))
                self.assertEqual(1, len(list(glob('.file.*.tmp.txt'))))
                raise Exception("test")
        except Exception as e:
            if e.args[0] != 'test':
                raise # raise any error but our test error
        # File *doesn't* exist
        self.assertFalse(os.path.exists('file.txt'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.txt'))))

        with pt.io.finalized_open('file.txt', 't') as f:
            f.write('Guess who\'s back')
            self.assertFalse(os.path.exists('file.txt'))
            self.assertEqual(1, len(list(glob('.file.*.tmp.txt'))))
        # File *does* exist
        self.assertTrue(os.path.exists('file.txt'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.txt'))))
        with open('file.txt', 'rt') as f:
            self.assertEqual(f.read(), 'Guess who\'s back')

        with pt.io.finalized_open('file.txt', 't') as f:
            f.write('Shady\'s back')
            self.assertTrue(os.path.exists('file.txt'))
            self.assertEqual(1, len(list(glob('.file.*.tmp.txt'))))
            with open('file.txt', 'rt') as f:
                self.assertEqual(f.read(), 'Guess who\'s back')
        # contents *are* updated
        self.assertTrue(os.path.exists('file.txt'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.txt'))))
        with open('file.txt', 'rt') as f:
            self.assertEqual(f.read(), 'Shady\'s back')

        try:
            with pt.io.finalized_open('file.txt', 't') as f:
                f.write('Back again')
                self.assertTrue(os.path.exists('file.txt'))
                self.assertEqual(1, len(list(glob('.file.*.tmp.txt'))))
                with open('file.txt', 'rt') as f:
                    self.assertEqual(f.read(), 'Shady\'s back')
                raise Exception("test")
        except Exception as e:
            if e.args[0] != 'test':
                raise # raise any error but our test error
        # contents *aren't* updated
        self.assertTrue(os.path.exists('file.txt'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.txt'))))
        with open('file.txt', 'rt') as f:
            self.assertEqual(f.read(), 'Shady\'s back')

    def test_finalized_autoopen(self):
        self.assertFalse(os.path.exists('file.txt'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.gz'))))

        try:
            with pt.io.finalized_autoopen('file.gz', 't') as f:
                f.write('OK')
                self.assertFalse(os.path.exists('file.gz'))
                self.assertEqual(1, len(list(glob('.file.*.tmp.gz'))))
                raise Exception("test")
        except Exception as e:
            if e.args[0] != 'test':
                raise # raise any error but our test error
        # File *doesn't* exist
        self.assertFalse(os.path.exists('file.gz'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.gz'))))

        with pt.io.finalized_autoopen('file.gz', 't') as f:
            f.write('Guess who\'s back')
            self.assertFalse(os.path.exists('file.gz'))
            self.assertEqual(1, len(list(glob('.file.*.tmp.gz'))))
        # File *does* exist
        self.assertTrue(os.path.exists('file.gz'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.gz'))))
        with pt.io.autoopen('file.gz', 'rt') as f:
            self.assertEqual(f.read(), 'Guess who\'s back')

        with pt.io.finalized_autoopen('file.gz', 't') as f:
            f.write('Shady\'s back')
            self.assertTrue(os.path.exists('file.gz'))
            self.assertEqual(1, len(list(glob('.file.*.tmp.gz'))))
            with pt.io.autoopen('file.gz', 'rt') as f:
                self.assertEqual(f.read(), 'Guess who\'s back')
        # contents *are* updated
        self.assertTrue(os.path.exists('file.gz'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.gz'))))
        with pt.io.autoopen('file.gz', 'rt') as f:
            self.assertEqual(f.read(), 'Shady\'s back')

        try:
            with pt.io.finalized_autoopen('file.gz', 't') as f:
                f.write('Back again')
                self.assertTrue(os.path.exists('file.gz'))
                self.assertEqual(1, len(list(glob('.file.*.tmp.gz'))))
                with pt.io.autoopen('file.gz', 'rt') as f:
                    self.assertEqual(f.read(), 'Shady\'s back')
                raise Exception("test")
        except Exception as e:
            if e.args[0] != 'test':
                raise # raise any error but our test error
        # contents *aren't* updated
        self.assertTrue(os.path.exists('file.gz'))
        self.assertEqual(0, len(list(glob('.file.*.tmp.gz'))))
        with pt.io.autoopen('file.gz', 'rt') as f:
            self.assertEqual(f.read(), 'Shady\'s back')

    def test_read_results_topic_merging(self):
        orig_results = pd.DataFrame({
            'qid': ['1', '1', '2'],
            'docno': ['A', 'B', 'C'],
            'score': [0.8, 0.4, 0.6],
            'rank': [1, 2, 1]
        })
        pt.io.write_results(orig_results, 'test.res')
        for results in [
            pt.io.read_results('test.res', dataset='vaswani'),
            pt.io.read_results('test.res', dataset=pt.get_dataset('vaswani')),
            pt.io.read_results('test.res', topics=pt.get_dataset('vaswani').get_topics()),]:
            self.assertEqual(results.iloc[0].query, 'MEASUREMENT OF DIELECTRIC CONSTANT OF LIQUIDS BY THE USE OF MICROWAVE TECHNIQUES')
            self.assertEqual(results.iloc[1].query, 'MEASUREMENT OF DIELECTRIC CONSTANT OF LIQUIDS BY THE USE OF MICROWAVE TECHNIQUES')
            self.assertEqual(results.iloc[2].query, 'MATHEMATICAL ANALYSIS AND DESIGN DETAILS OF WAVEGUIDE FED MICROWAVE RADIATIONS')

    def tearDown(self):
        for file in ['file.txt', 'file.tmp.txt', 'file.gz', 'file.tmp.gz', 'test.res']:
            if os.path.exists(file):
                os.remove(file)

    

if __name__ == "__main__":
    unittest.main()



from pyterrier.datasets._core import Dataset, DatasetProvider, get_dataset, find_datasets, list_datasets, transformer_from_dataset, DATASET_MAP, RemoteDataset

# _builtin and _irds is not loaded here -- they are loaded on-demand once a dataset is requested that needs it.

__all__ = ["Dataset", "DatasetProvider", "get_dataset", "find_datasets", "list_datasets", "transformer_from_dataset", "DATASET_MAP", "RemoteDataset"]



import pandas as pd
import pyterrier as pt
from .base import BaseTestCase, parallel_test


class TestGrid(BaseTestCase):

    def test_gridscan_2pipe_2params(self):
        dataset = pt.get_dataset("vaswani")
        index = pt.IndexFactory.of(dataset.get_index())
        PL2 = pt.terrier.Retriever(index, wmodel="PL2", controls={'c' : 1})
        bo1 = pt.rewrite.Bo1QueryExpansion(index)
        pipe = PL2 >> bo1 >> PL2                

        rtr = pt.GridScan(
            pipe, 
            {
                PL2 : {'c' : [0.1, 1]},
                bo1 : {'fb_terms' : [2,4]}}, 
            dataset.get_topics().head(2),
            dataset.get_qrels(),
            dataframe=False
        )
        self.assertEqual(4, len(rtr))
        #print(rtr)

    def test_gridscan_2params(self):
        dataset = pt.get_dataset("vaswani")
        pipe = pt.terrier.Retriever(dataset.get_index(), wmodel="PL2", controls={'c' : 1})
        self.assertEqual(1, pipe.get_parameter('c'))
        self.assertEqual("PL2", pipe.get_parameter('wmodel'))
        rtr = pt.GridScan(
            pipe, 
            {pipe : {'c' : [0.1, 1], 'wmodel' : ["PL2", "BM25"]}}, 
            dataset.get_topics().head(2),
            dataset.get_qrels(),
            dataframe=False
        )
        self.assertEqual(4, len(rtr))
        #print(rtr)

    def test_gridscan_1param(self):
        dataset = pt.get_dataset("vaswani")
        pipe = pt.terrier.Retriever(dataset.get_index(), wmodel="PL2", controls={'c' : 1})
        self.assertEqual(1, pipe.get_parameter('c'))
        rtr = pt.GridScan(
            pipe, 
            {pipe : {'c' : [0.1, 1, 5, 10, 20, 100]}}, 
            dataset.get_topics().head(5),
            dataset.get_qrels(),
            dataframe=False
        )
        self.assertEqual(6, len(rtr))
        #print(rtr)
    
    def test_gridscan_1param_df(self):
        dataset = pt.get_dataset("vaswani")
        pipe = pt.terrier.Retriever(dataset.get_index(), wmodel="PL2", controls={'c' : 1})
        self.assertEqual(1, pipe.get_parameter('c'))
        rtr = pt.GridScan(
            pipe, 
            {pipe : {'c' : [0.1, 1, 5, 10, 20, 100]}}, 
            dataset.get_topics().head(5),
            dataset.get_qrels(),
            dataframe=True,
        )
        self.assertEqual(6, len(rtr))
        self.assertTrue(isinstance(rtr, pd.DataFrame))
        #print(rtr)

    @parallel_test
    def test_gridscan_joblib2(self):
        self.skip_windows()
        dataset = pt.get_dataset("vaswani")
        pipe = pt.terrier.Retriever(dataset.get_index(), wmodel="PL2", controls={'c' : 1})
        self.assertEqual(1, pipe.get_parameter('c'))
        rtr = pt.GridScan(
            pipe, 
            {pipe : {'c' : [0.1, 1, 5, 10, 20, 100]}}, 
            dataset.get_topics().head(5),
            dataset.get_qrels(),
            jobs=2,
            dataframe=False
        )
        self.assertEqual(6, len(rtr))
        #print(rtr)

    def test_gridsearch(self):
        dataset = pt.get_dataset("vaswani")
        pipe = pt.terrier.Retriever(dataset.get_index(), wmodel="PL2", controls={'c' : 1})
        rtr = pt.GridSearch(
            pipe, 
            {pipe : {'c' : [0.1, 1, 5, 10, 20, 100]}}, 
            dataset.get_topics().head(5),
            dataset.get_qrels()
        )
        self.assertEqual(100, rtr.get_parameter("c"))

    def test_kfoldgridsearch(self):
        import pandas as pd
        dataset = pt.get_dataset("vaswani")
        topics = dataset.get_topics().head(10)
        topics_part = [
            topics.iloc[0:1],
            topics.iloc[2:3],
            topics.iloc[4:5],
            topics.iloc[6:7],
            topics.iloc[8:9]           
        ]
        pipe = pt.terrier.Retriever(dataset.get_index(), wmodel="PL2", controls={'c' : 1})
        rtrDf, rtrSettings = pt.KFoldGridSearch(
            pipe, 
            {pipe : {'c' : [0.1, 1, 5, 10, 20, 100]}},
            topics_part,
            dataset.get_qrels()
        )
        self.assertTrue(isinstance(rtrDf, pd.DataFrame))
        #print(rtrSettings)


import pandas as pd
import pyterrier as pt
from .base import *

class TestFromDf(BaseTestCase):

    def test_1(self):
        queries = pt.new.queries(["this is a query"])
        docs = pt.new.ranked_documents([[1, 2]], qid=["1"])
        rtr = pt.transformer.SourceTransformer(docs).transform(queries)
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("score" in rtr.columns)
        self.assertTrue("docno" in rtr.columns)
        self.assertEqual(2, len(rtr))

    def test_1_replace_query(self):
        queries = pt.new.queries(["this is a query"])
        docs = pt.new.ranked_documents([[1, 2]], qid=["1"], query=[["new query", "new query"]])
        rtr = pt.transformer.SourceTransformer(docs).transform(queries)
        self.assertEqual(2, len(rtr))
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("score" in rtr.columns)
        self.assertTrue("docno" in rtr.columns)
        self.assertEqual("new query", rtr.iloc[0].query)
        self.assertEqual("new query", rtr.iloc[1].query)        

    def test_query_cols(self):
        queries = pt.new.queries(["this is a query"], query_embs=[[1,1]])
        self.assertTrue("query_embs" in queries.columns)
        docs = pt.new.ranked_documents([[1, 2]], qid=["1"])
        rtr = pt.transformer.SourceTransformer(docs).transform(queries)
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("score" in rtr.columns)
        self.assertTrue("docno" in rtr.columns)
        self.assertTrue("query_embs" in rtr.columns)
        self.assertEqual(2, len(rtr))



from enum import Enum
import pyterrier as pt

class TerrierTokeniser(Enum):
    """A built-in Terrier tokeniser.

    This enum is primarily used with :class:`~pyterrier.terrier.TerrierIndexer.indexer`.
    """

    whitespace = 'whitespace' #: Tokenise on whitespace only
    english = 'english' #: Terrier's standard tokeniser, designed for English
    utf = 'utf' #: A variant of Terrier's standard tokeniser, similar to English, but with UTF support.
    twitter = 'twitter' #: Like utf, but keeps hashtags etc
    identity = 'identity' #: Performs no tokenisation - strings are kept as is. 

    @staticmethod
    def _to_obj(this) -> 'TerrierTokeniser':
        try:
            return TerrierTokeniser(this)
        except ValueError:
            return this

    @staticmethod
    def _to_class(this) -> str:
        if this == TerrierTokeniser.whitespace:
            return 'WhitespaceTokeniser'
        if this == TerrierTokeniser.english:
            return 'EnglishTokeniser'
        if this == TerrierTokeniser.utf:
            return 'UTFTokeniser'
        if this == TerrierTokeniser.twitter:
            return 'UTFTwitterTokeniser'
        if this == TerrierTokeniser.identity:
            return 'IdentityTokeniser'
        if isinstance(this, str):
            return this
        raise ValueError(f'Unknown/unsupported tokeniser: {this}')
    
    @staticmethod
    @pt.java.required
    def java_tokeniser(this):
        clz = TerrierTokeniser._to_class(this)
        if "." not in clz:
            clz = "org.terrier.indexing.tokenisation." + clz
        return pt.java.autoclass(clz)()
